#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.errors import ClientError
from datarobot.models.api_object import APIObject
from datarobot.models.deployment import Deployment
from datarobot.models.genai.playground import Playground
from datarobot.models.model import Model
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id, resolve_use_cases, UseCaseLike
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


def get_entity_id(entity: Union[CustomModelValidation, Deployment, Model, Playground, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : APIObject or str
        Specifies either the entity ID or the entity.

    Returns
    -------
    id : str
        The entity ID.
    """
    return entity if isinstance(entity, str) else entity.id


custom_model_validation_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
        t.Key("prompt_column_name"): t.String,
        t.Key("target_column_name"): t.String,
        t.Key("deployment_id"): t.String,
        t.Key("validation_status"): t.String,
        t.Key("model_id"): t.String,
        t.Key("deployment_access_data", optional=True, default=None): t.Or(
            t.Null,
            t.Dict(
                {
                    t.Key("prediction_api_url"): t.String,
                    t.Key("datarobot_key", optional=True, default=None): t.Or(t.Null, t.String),
                    t.Key("authorization_header"): t.String,
                    t.Key("input_type"): t.String,
                    t.Key("model_type"): t.String,
                }
            ).ignore_extra("*"),
        ),
        t.Key("tenant_id"): t.String,
        t.Key("user_id"): t.String,
        t.Key("creation_date"): t.String,
        t.Key("error_message", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("deployment_name", optional=True, default=None): t.Or(
            t.Null, t.String(allow_blank=True)
        ),
        t.Key("user_name", optional=True, default=None): t.Or(t.Null, t.String(allow_blank=True)),
        t.Key("use_case_id", optional=True, default=None): t.Or(t.Null, t.String),
        t.Key("prediction_timeout"): t.Int,
    }
).ignore_extra("*")


class CustomModelValidation(APIObject):
    """
    The validation record checking the ability of the deployment to serve
    as a custom model LLM, custom model vector database, or custom model embedding.

    Attributes
    ----------
    id : str
        The ID of the validation.
    prompt_column_name : str
        The name of the column the deployed model uses for prompt text input.
    target_column_name : str
        The name of the column the deployed model uses for prediction output.
    deployment_id : str
        The ID of the deployment.
    model_id : str
        The ID of the underlying deployed model, which can be found using `Deployment.model["id"]`.
    validation_status : str
        Can be TESTING, FAILED, or PASSED. Only PASSED is allowed for use.
    deployment_access_data : dict, optional
        The data that will be used for accessing the deployment prediction server.
        This field is only available for deployments that pass validation.
        Dict fields are as follows:
        - prediction_api_url - The URL for the deployment prediction server.
        - datarobot_key - The first of two auth headers for the prediction server.
        - authorization_header - The second of two auth headers for the prediction server.
        - input_type - The input type the model expects, either JSON or CSV.
        - model_type - The target type of the deployed custom model.
    tenant_id : str
        The creating user's tenant ID.
    name : str
        The display name of the validated custom model.
    creation_date : str
        The creation date of the validation (ISO 8601 formatted).
    user_id : str
        The ID of the creating user.
    error_message : Optional[str]
        Additional information for the errored validation.
    deployment_name : Optional[str]
        The name of the validated deployment.
    user_name : Optional[str]
        The name of the creating user.
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    prediction_timeout: int
        The timeout, in seconds, for the prediction API used in this custom model validation.
    """

    _path: str
    _converter = custom_model_validation_trafaret

    def __init__(
        self,
        id: str,
        prompt_column_name: str,
        target_column_name: str,
        deployment_id: str,
        model_id: str,
        validation_status: str,
        deployment_access_data: Optional[Dict[str, Any]],
        tenant_id: str,
        name: str,
        creation_date: str,
        user_id: str,
        error_message: Optional[str],
        deployment_name: Optional[str],
        user_name: Optional[str],
        use_case_id: Optional[str],
        prediction_timeout: int,
    ):
        self.id = id
        self.prompt_column_name = prompt_column_name
        self.target_column_name = target_column_name
        self.deployment_id = deployment_id
        self.model_id = model_id
        self.validation_status = validation_status
        self.deployment_access_data = deployment_access_data
        self.tenant_id = tenant_id
        self.error_message = error_message
        self.name = name
        self.creation_date = creation_date
        self.user_id = user_id
        self.deployment_name = deployment_name
        self.user_name = user_name
        self.use_case_id = use_case_id
        self.prediction_timeout = prediction_timeout

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id})"

    @classmethod
    def get(cls, validation_id: str) -> CustomModelValidation:
        """
        Get the validation record by id.

        Parameters
        ----------
        validation_id : Union[CustomModelValidation, str]
            The CustomModelValidation to retrieve, either `CustomModelValidation` or validation ID.

        Returns
        -------
        CustomModelValidation
        """
        url = f"{cls._client.domain}/{cls._path}/{get_entity_id(validation_id)}/"
        response = cls._client.get(url)
        return cls.from_server_data(response.json())

    @classmethod
    def get_by_values(
        cls, prompt_column_name: str, target_column_name: str, deployment_id: str, model_id: str
    ) -> CustomModelValidation:
        """
        Get the validation record by field values.

        Parameters
        ----------
        prompt_column_name : str
            The name of the column the deployed model uses for prompt text input.
        target_column_name : str
            The name of the column the deployed model uses for prediction output.
        deployment_id : str
            The ID of the deployment.
        model_id : str
            The ID of the underlying deployed model.

        Returns
        -------
        CustomModelValidation
        """

        url = f"{cls._client.domain}/{cls._path}/"
        params = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": deployment_id,
            "model_id": model_id,
            "order_by": "-creationDate",
            "limit": 1,
        }
        response_body = cls._client.get(url, params=params).json()
        data = response_body.get("data")

        if data is None:
            return cls.from_server_data(response_body)
        else:
            if len(data) == 0:
                raise ClientError(
                    exc_message="Custom model LLM validation not found", status_code=404
                )
            else:
                return cls.from_server_data(data[0])

    @classmethod
    def list(
        cls,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        deployment: Optional[Union[Deployment, str]] = None,
        model: Optional[Union[Model, str]] = None,
        use_cases: Optional[UseCaseLike] = None,
        playground: Optional[Union[Playground, str]] = None,
        completed_only: bool = False,
        search: Optional[str] = None,
        sort: Optional[str] = None,
    ) -> List[CustomModelValidation]:
        """
        List the validation records by field values.

        Parameters
        ----------
        prompt_column_name : Optional[str], optional
            The column name the deployed model expects as the input.
        target_column_name : Optional[str], optional
            The target name that the deployed model will output.
        deployment : Optional[Union[Deployment, str]], optional
            The returned validations are filtered to those associated with a specific deployment
            if specified, either `Deployment` or the deployment ID.
        model_id : Optional[Union[Model, str]], optional
            The returned validations are filtered to those associated with a specific model
            if specified, either `Model` or model ID.
        use_cases : Optional[list[Union[UseCase, str]]], optional
            The returned validations are filtered to those associated with specific Use Cases
            if specified, either `UseCase` objects or the Use Case IDs.
        playground_id : Optional[Union[Playground, str]], optional
            The returned validations are filtered to those used in a specific playground
            if specified, either `Playground` or playground ID.
        completed_only : Optional[bool]
            Whether to retrieve only completed validations.
        search : Optional[str], optional
            String for filtering validations.
            Validations that contain the string in name will be returned.
        sort : Optional[str], optional
            Property to sort validations by.
            Prefix the attribute name with a dash to sort in descending order,
            e.g. sort='-name'.
            Currently supported options are listed in ListCustomModelValidationsSortQueryParams
            but the values can differ with different platform versions.
            By default, the sort parameter is None which will result in
            validations being returned in order of creation time descending.

        Returns
        -------
        List[CustomModelValidation]
        """

        url = f"{cls._client.domain}/{cls._path}/"
        params = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": get_entity_id(deployment) if deployment else None,
            "model_id": get_entity_id(model) if model else None,
            "playground_id": get_entity_id(playground) if playground else None,
            "completed_only": completed_only,
            "search": search,
            "sort": sort if sort else "-creationDate",
        }
        params = resolve_use_cases(use_cases=use_cases, params=params, use_case_key="use_case_id")
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    @classmethod
    def revalidate(cls, validation_id: str) -> CustomModelValidation:
        """
        Revalidate an unlinked custom model vector database or LLM.
        This method is helpful when a deployment used as vector database or LLM is accidentally
        replaced with another model that stopped complying with the response schema requirements.

        Replace the deployment's model with a complying model and call this method instead of
        creating a new custom model validation from scratch.

        Another application is if the API token used to create a validation record got revoked and
        no longer can be used to call the deployment.
        Calling `revalidate` will update the validation record with the token currently in use.

        Parameters
        ----------
        validation_id : str
            The ID of the CustomModelValidation for revalidation.

        Returns
        -------
        CustomModelValidation
        """
        url = f"{cls._client.domain}/{cls._path}/{validation_id}/revalidate/"
        response = cls._client.post(url)
        return cls.from_server_data(response.json())

    def delete(self) -> None:
        """
        Delete the custom model validation.
        """
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        self._client.delete(url)


class NonChatAwareCustomModelValidation(CustomModelValidation):
    """
    The validation record checking the ability of the deployment to serve as a custom model resource
    (either vector database or an embedding model), while not using the chat completions API for
    inference.

    Attributes
    ----------
    id : str
        The ID of the validation.
    prompt_column_name : str
        The name of the column the deployed model uses for prompt text input.
    target_column_name : str
        The name of the column the deployed model uses for prediction output.
    deployment_id : str
        The ID of the deployment.
    model_id : str
        The ID of the underlying deployed model, which can be found using `Deployment.model["id"]`.
    validation_status : str
        Can be TESTING, FAILED and PASSED. Only PASSED allowed for use.
    deployment_access_data : dict, optional
        The data that will be used for accessing the deployment prediction server.
        This field is only available for deployments that pass validation.
        Dict fields are as follows:
        - prediction_api_url - The URL for the deployment prediction server.
        - datarobot_key - The first of two auth headers for the prediction server.
        - authorization_header - The second of two auth headers for the prediction server.
        - input_type - The input type the model expects, either JSON or CSV.
        - model_type - The target type of the deployed custom model.
    tenant_id : str
        The creating user's tenant ID.
    name : str
        The display name of the validated custom model.
    creation_date : str
        The creation date of the validation (ISO 8601 formatted).
    user_id : str
        The ID of the creating user.
    error_message : Optional[str]
        Additional information for the errored validation.
    deployment_name : Optional[str]
        The name of the validated deployment.
    user_name : Optional[str]
        The name of the creating user.
    use_case_id : Optional[str]
        The ID of the Use Case associated with the validation.
    """

    @classmethod
    def create(
        cls,
        prompt_column_name: str,
        target_column_name: str,
        deployment_id: Union[Deployment, str],
        model: Optional[Union[Model, str]] = None,
        use_case: Optional[Union[UseCase, str]] = None,
        name: Optional[str] = None,
        wait_for_completion: bool = False,
        prediction_timeout: Optional[int] = None,
    ) -> NonChatAwareCustomModelValidation:
        """
        Start the validation of the deployment that will serve as a custom model resource
        (vector database, embedding model, etc.).

        Parameters
        ----------
        prompt_column_name : str
            The column name the deployed model expect as the input.
        target_column_name : str
            The target name that the deployed model will output.
        deployment_id : Union[Deployment, str]
            The deployment to validate, either `Deployment` or the deployment ID.
        model : Optional[Union[Model, str]], optional
            The specific model within the deployment, either `Model` or the model ID.
            If not specified, the underlying model ID will be automatically derived from the
            deployment information.
        use_case : Optional[Union[UseCase, str]], optional
            The Use Case to link the validation to, either `UseCase` or the Use Case ID.
        name : Optional[str], optional
            The name of the validation.
        wait_for_completion : bool
            If set to `True`, the code will wait for the validation job to complete before returning
            results. If the job does not finish in 10 minutes, this method call raises a timeout
            error.
            If set to `False`, the code does not wait for the job to complete. Instead,
            `CustomModelValidation.get` can be used to poll for the status of the job using
            the validation ID returned by the method.
        prediction_timeout : Optional[int], optional
            The timeout, in seconds, for the prediction API used in this custom model validation.

        Returns
        -------
        NonChatAwareCustomModelValidation
        """

        payload = {
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": get_entity_id(deployment_id),
            "model_id": get_entity_id(model) if model else None,
            "use_case_id": get_use_case_id(use_case, is_required=False),
            "name": name,
        }
        if prediction_timeout is not None:
            payload["prediction_timeout"] = prediction_timeout  # type: ignore[assignment]
        url = f"{cls._client.domain}/{cls._path}/"
        response = cls._client.post(url, data=payload)
        if wait_for_completion:
            location = wait_for_async_resolution(cls._client, response.headers["Location"])
            return cls.from_location(location)
        return cls.from_server_data(response.json())

    def update(
        self,
        name: Optional[str] = None,
        prompt_column_name: Optional[str] = None,
        target_column_name: Optional[str] = None,
        deployment: Optional[Union[Deployment, str]] = None,
        model: Optional[Union[Model, str]] = None,
        prediction_timeout: Optional[int] = None,
    ) -> NonChatAwareCustomModelValidation:
        """
        Update a custom model validation.

        Parameters
        ----------
        name : Optional[str], optional
            The new name of the custom model validation.
        prompt_column_name : Optional[str], optional
            The new name of the prompt column.
        target_column_name : Optional[str], optional
            The new name of the target column.
        deployment : Optional[Union[Deployment, str]], optional
            The new deployment to validate.
        model : Optional[Union[Model, str]], optional
            The new model within the deployment to validate.
        prediction_timeout : Optional[int], optional
            The new timeout, in seconds, for the prediction API used in this custom model validation.

        Returns
        -------
        NonChatAwareCustomModelValidation
        """
        payload = {
            "name": name,
            "prompt_column_name": prompt_column_name,
            "target_column_name": target_column_name,
            "deployment_id": get_entity_id(deployment) if deployment else None,
            "model_id": get_entity_id(model) if model else None,
            "prediction_timeout": prediction_timeout,
        }
        url = f"{self._client.domain}/{self._path}/{self.id}/"
        response = self._client.patch(url, data=payload)
        return self.from_server_data(response.json())



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime
from typing import Any, Dict, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.mlops.enums import EventStatusType, RemoteEventType
from datarobot.models.api_object import APIObject
from datarobot.utils import datetime_to_string

EVENT_MESSAGE_MAX_LENGTH = 16384
PE_LABEL_MAX_LENGTH = 512
MAX_GUARD_NAME_LENGTH = 512


class MLOpsEvent(APIObject):
    """
    An MLOps Event Object: An object representing an important MLOps activity that
    happened.  For example, health, service issues with the DataRobot deployment
    or a prediction environment or a particular phase in a long operation (like
    creation of deployment or processing training data) is completed or errored.

    This class allows the client to report such event to the DataRobot service.

    Notes
    -----
        DataRobot backend support lots of events and all these events are categorized
        into different categories.  This class does not yet support *ALL* events, but
        we will gradually add support for them

        Supported Event Categories:
            - moderation
    """

    _path = "remoteEvents/"

    _moderation_data = t.Dict(
        {
            t.Key("guard_name", optional=True): String(
                max_length=MAX_GUARD_NAME_LENGTH, allow_blank=True
            ),
            t.Key("metric_name", optional=True): String(
                max_length=MAX_GUARD_NAME_LENGTH, allow_blank=True
            ),
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("title", optional=True): String(max_length=PE_LABEL_MAX_LENGTH),
            t.Key("status_type", optional=True): t.Enum(*[e.value for e in EventStatusType]),
            t.Key("type"): t.Enum(*[e.value for e in RemoteEventType]),
            t.Key("timestamp"): String(),
            t.Key("message", optional=True): String(max_length=EVENT_MESSAGE_MAX_LENGTH),
            t.Key("moderation_data", optional=True): _moderation_data,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        type: RemoteEventType,
        status_type: EventStatusType,
        timestamp: Optional[str] = None,
        title: Optional[str] = None,
        message: Optional[str] = None,
        moderation_data: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.id = id
        self.type = type
        self.title = title
        self.message = message
        self.moderation_data = moderation_data
        self.timestamp = timestamp
        self.status_type = status_type

    @classmethod
    def report_moderation_event(
        cls,
        event_type: str,
        timestamp: Optional[str] = None,
        title: Optional[str] = None,
        message: Optional[str] = None,
        deployment_id: Optional[str] = None,
        org_id: Optional[str] = None,
        guard_name: Optional[str] = None,
        metric_name: Optional[str] = None,
    ) -> None:
        """
        Reports a moderation event

        Parameters
        ----------
        event_type : str
            The type of the moderation event.
        timestamp : Optional[str]
            The timestamp of the event, datetime, or string in RFC3339 format. If the datetime provided
            does not have a timezone, DataRobot assumes it is UTC.
        title : Optional[str]
            The title of the moderation event.
        message : Optional[str]
            A description of the moderation event.
        deployment_id : Optional[str]
            The ID of the deployment associated with the event.
        org_id : Optional[str]
            The ID of the organization associated with the event.
        guard_name : Optional[str]
            The name or label of the guard.
        metric_name : Optional[str]
            The name or label of the metric.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If event_type is not one of the moderation event types.
            If fails to create the event.

        Examples
        --------
        .. code-block:: python

            >>> from datarobot.mlops.events import MLOpsEvent
            >>> MLOpsEvent.report_moderation_event(
            ...     event_type="moderationMetricCreationError",
            ...     title="Failed to create moderation metric",
            ...     message="Maximum number of custom metrics reached",
            ...     deployment_id="5c939e08962d741e34f609f0",
            ...     metric_name="Blocked Prompts",
            ... )
        """
        if event_type not in RemoteEventType.moderation_event_types():
            raise ValueError(
                f"Invalid moderation event type {event_type}, supported events are: "
                f"{RemoteEventType.moderation_event_types()}"
            )

        _timestamp = timestamp
        if _timestamp is None:
            _timestamp = datetime_to_string(datetime.utcnow())

        payload = {
            "title": title,
            "message": message,
            "timestamp": _timestamp,
            "deploymentId": str(deployment_id),
            "eventType": event_type,
            "moderationData": {
                "guardName": guard_name if guard_name else "",
                "metricName": metric_name if metric_name else "",
            },
            "orgId": org_id,
        }

        try:
            cls._client.post(cls._path, data=payload)
        except Exception as ex:
            raise Exception(f"Failed to create moderation event: {ex}")



#
# Copyright 2025 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import json
from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.enums import ListCustomTemplatesSortQueryParams
from datarobot.models.api_object import APIObject
from datarobot.utils import _from_api_dict, to_api


class DefaultEnvironment(APIObject):
    """
    Default execution environment.
    """

    _converter = t.Dict(
        {
            t.Key("environment_id"): t.String(),
            t.Key("environment_version_id"): t.String(),
        }
    ).ignore_extra("*")
    schema = _converter

    def __init__(self, environment_id: str, environment_version_id: str):
        self.environment_id = environment_id
        self.environment_version_id = environment_version_id

    def __repr__(self) -> str:
        return "{}(environment_id={}, environment_version_id={})".format(
            self.__class__.__name__, self.environment_id, self.environment_version_id
        )


class CustomMetricMetadata(APIObject):
    """
    Metadata for custom metrics.
    """

    _converter = t.Dict(
        {
            # NOTE: several values are really enums, but treated as strings for simplicity
            t.Key("units"): t.String(),
            t.Key("directionality"): t.String(),
            t.Key("type"): t.String(),
            t.Key("time_step"): t.String(),
            t.Key("is_model_specific"): t.Bool(),
            t.Key("template_metric_type", optional=True): t.Or(t.Null(), t.String()),
        }
    ).ignore_extra("*")
    schema = _converter

    def __init__(
        self,
        units: str,
        directionality: str,
        type: str,
        time_step: str,
        is_model_specific: bool,
        template_metric_type: Optional[str] = None,
    ):
        self.units = units
        self.directionality = directionality
        self.type = type
        self.time_step = time_step
        self.is_model_specific = is_model_specific
        self.template_metric_type = template_metric_type


class TemplateMetadata(APIObject):
    """
    Metadata for the custom templates.
    """

    _converter = t.Dict(
        {
            t.Key("readme", optional=True): t.Or(t.Null(), t.String()),
            t.Key("source", optional=True): t.Or(t.Null(), t.Dict().allow_extra("*")),
            t.Key("tags", optional=True): t.List(t.String()),
            t.Key("custom_metric_metadata", optional=True): t.Or(
                t.Null(), CustomMetricMetadata.schema
            ),
            t.Key("feature_flag", optional=True): t.Or(t.Null(), t.String()),
            t.Key("preview_image", optional=True): t.Or(t.Null(), t.String()),
        }
    ).ignore_extra("*")
    schema = _converter

    def __init__(
        self,
        readme: Optional[str] = None,
        source: Optional[str] = None,
        tags: Optional[List[str]] = None,
        feature_flag: Optional[str] = None,
        preview_image: Optional[str] = None,
    ):
        self.readme = readme
        self.source = source
        self.tags = tags
        self.feature_flag = feature_flag
        self.preview_image = preview_image


class CustomTemplate(APIObject):
    """
    Template for custom activity (e.g. custom-metrics, applications).
    """

    _path = "customTemplates/"

    _converter = t.Dict(
        {
            t.Key("default_environment"): DefaultEnvironment.schema,
            t.Key("default_resource_bundle_id", optional=True, default=None): t.Or(
                t.Null(), t.String()
            ),
            t.Key("description"): t.String(),
            t.Key("enabled"): t.Bool(),
            t.Key("id"): t.String(),
            t.Key("items"): t.List(t.Dict().allow_extra("*")),
            t.Key("name"): t.String(),
            t.Key("template_metadata"): TemplateMetadata.schema,
            t.Key("template_sub_type"): t.String(),
            t.Key("template_type"): t.String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        template_metadata: Dict[str, Any],
        default_environment: Dict[str, Any],
        default_resource_bundle_id: Optional[str],
        items: List[Dict[str, Any]],
        template_type: str,
        template_sub_type: str,
        enabled: bool,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.template_metadata = template_metadata
        self.default_environment = default_environment
        self.default_resource_bundle_id = default_resource_bundle_id
        self.items = items
        self.template_type = template_type
        self.template_sub_type = template_sub_type
        self.enabled = enabled

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id!r}, name={self.name!r})"

    @classmethod
    def list(
        cls,
        search: Optional[str] = None,
        order_by: Optional[ListCustomTemplatesSortQueryParams] = None,
        tag: Optional[str] = None,
        template_type: Optional[str] = None,
        template_sub_type: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[CustomTemplate]:
        """List all custom templates.

        .. versionadded:: v3.7

        Parameters
        ----------
        search: Optional[str]
            Search string.
        order_by: Optional[ListCustomTemplatesSortQueryParams]
            Ordering field.
        tag: Optional[str]
            Tag associated with the template.
        template_type: Optional[str]
            Type of the template.
        template_type: Optional[str]
            Sub-type of the template.
        offset: Optional[int]
            Offset for pagination.
        limit: Optional[int]
            Limit for pagination.

        Returns
        -------
        templates: List[CustomTemplate]
        """
        params: Dict[str, Any] = {}
        if search:
            params["search"] = search
        if order_by:
            params["orderBy"] = order_by
        if tag:
            params["tag"] = tag
        if template_type:
            params["templateType"] = template_type
        if template_sub_type:
            params["templateSubType"] = template_sub_type
        if offset:
            params["offset"] = offset
        if limit:
            params["limit"] = limit
        response = cls._client.get(cls._path, params=params if params else None)
        return [cls.from_server_data(d) for d in response.json()["data"]]

    @classmethod
    def get(cls, template_id: str) -> CustomTemplate:
        """Get a custom template by ID.

        .. versionadded:: v3.7

        Parameters
        ----------
        template_id: str
            ID of the template.

        Returns
        -------
        template : CustomTemplate
        """
        response = cls._client.get(f"{cls._path}{template_id}/")
        return cls.from_server_data(response.json())

    def update(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        default_resource_bundle_id: Optional[str] = None,
        template_type: Optional[str] = None,
        template_sub_type: Optional[str] = None,
        template_metadata: Optional[TemplateMetadata | str] = None,
        default_environment: Optional[DefaultEnvironment | str] = None,
    ) -> None:
        """
        Update the custom template.

        .. versionadded:: v3.7

        Parameters
        ----------
        name: Optional[str]
            The template name.
        description: Optional[str]
            A description of the template.
        default_resource_bundle_id: Optional[str]
            The default resource bundle ID.
        template_type: Optional[str]
            The template type.
        template_sub_type: Optional[str]
            The template sub-type.
        template_metadata: Optional[TemplateMetadata|str]
            The metadata associated with the template, provided as TemplateMetadata or a JSON encoded string.
        default_environment: Optional[DefaultEnvironment|str]
            The default environment associated with the template, provided as DefaultEnvironment or a JSON encoded
            string.

        Examples
        --------
        .. code-block:: python

            from datarobot import CustomTemplate
            from datarobot.models.custom_templates import DefaultEnvironment
            new_env = DefaultEnvironment(
                environment_id='679d47c8ce1ecd17326f3fdf',
                environment_version_id='679d47c8ce1ecd17326f3fe3',
            )
            template = CustomTemplate.get(template_id='5c939e08962d741e34f609f0')
            template.update(default_environment=new_env, description='Updated template with environment v17')
        """

        _default_env_obj: Any = None
        _template_meta_obj: Any = None
        body = {}
        if name:
            body["name"] = name
        if description:
            body["description"] = description
        if default_resource_bundle_id:
            body["defaultResourceBundleId"] = default_resource_bundle_id
        if template_type:
            body["templateType"] = template_type
        if template_sub_type:
            body["templateSubType"] = template_sub_type
        if template_metadata:
            if isinstance(template_metadata, TemplateMetadata):
                api_obj = to_api(template_metadata)
                _template_meta_obj = _from_api_dict(api_obj)  # type: ignore[arg-type]
                template_metadata = json.dumps(api_obj)
            else:
                _template_meta_obj = _from_api_dict(json.loads(template_metadata))
            body["templateMetadata"] = template_metadata
        if default_environment:
            if isinstance(default_environment, DefaultEnvironment):
                api_obj = to_api(default_environment)
                _default_env_obj = _from_api_dict(api_obj)  # type: ignore[arg-type]
                default_environment = json.dumps(api_obj)
            else:
                _default_env_obj = _from_api_dict(json.loads(default_environment))
            body["defaultEnvironment"] = default_environment
        if not body:
            raise ValueError("Nothing to update")

        url = f"{self._path}{self.id}/"
        self._client.patch(url, data=body)

        # update the local data
        if name:
            self.name = name
        if description:
            self.description = description
        if default_resource_bundle_id:
            self.default_resource_bundle_id = default_resource_bundle_id
        if template_type:
            self.template_type = template_type
        if template_sub_type:
            self.template_sub_type = template_sub_type
        if _template_meta_obj:
            self.template_metadata = _template_meta_obj
        if _default_env_obj:
            self.default_environment = _default_env_obj

    def delete(self) -> None:
        """
        Delete this custom template.

        .. versionadded:: v3.7

        Returns
        -------
        None
        """

        url = f"{self._path}{self.id}/"
        self._client.delete(url)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.enums import AggregationType, enum_to_list
from datarobot.models.api_object import APIObject
from datarobot.models.genai.insights_configuration import InsightsConfiguration
from datarobot.utils.pagination import unpaginate

aggregation_value_trafaret = t.Dict(
    {
        t.Key("item"): t.String,
        t.Key("value"): t.Float,
    }
).ignore_extra("*")

evaluation_dataset_metric_aggregation_trafaret = t.Dict(
    {
        t.Key("llm_blueprint_id"): t.String,
        t.Key("evaluation_dataset_configuration_id"): t.String,
        t.Key("ootb_dataset_name"): t.Or(t.String, t.Null),
        t.Key("metric_name"): t.String(allow_blank=True),
        t.Key("deployment_id"): t.Or(t.String, t.Null),
        t.Key("dataset_id"): t.Or(t.String, t.Null),
        t.Key("dataset_name"): t.Or(t.String(allow_blank=True), t.Null),
        t.Key("chat_id"): t.String,
        t.Key("chat_name"): t.String(allow_blank=True),
        t.Key("aggregation_value"): t.Or(t.Float, t.List(aggregation_value_trafaret)),
        t.Key("aggregation_type"): t.Enum(*enum_to_list(AggregationType)),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("tenant_id"): t.String,
    }
).ignore_extra("*")


class EvaluationDatasetMetricAggregation(APIObject):
    """Information about an evaluation dataset metric aggregation job.
      This job runs a metric against LLMs using an evaluation dataset and aggregates the results.

    Attributes
    ----------
    llm_blueprint_id : str
        The LLM blueprint ID.
    evaluation_dataset_configuration_id : str
        The evaluation dataset configuration ID.
    ootb_dataset_name : str | None
        The name of the Datarobot-provided dataset that does not require additional configuration..
    metric_name : str
        The name of the metric.
    deployment_id : str | None
        A deployment ID if the evaluation was run against a deployment.
    dataset_id : str | None
        The ID of the dataset used in the evaluation.
    dataset_name : str | None
        The name of the dataset used in the evaluation.
    chat_id : str
        The ID of the chat created to run the evaluation.
    chat_name : str
        The name of the chat that was created to run the evaluation.
    aggregation_value : float | List[Dict[str, float]]
        The aggregated metric result.
    aggregation_type : AggregationType
        The type of aggregation used for the metric results.
    creation_date : str
        The date the evaluation job was created.
    creation_user_id : str
        The ID of the user who created the evaluation job.
    tenant_id : str
        The ID of the tenant that owns the evaluation job.
    """

    _path = "api/v2/genai/evaluationDatasetMetricAggregations"
    _converter = evaluation_dataset_metric_aggregation_trafaret

    def __init__(
        self,
        llm_blueprint_id: str,
        evaluation_dataset_configuration_id: str,
        ootb_dataset_name: Optional[str],
        metric_name: str,
        deployment_id: Optional[str],
        dataset_id: Optional[str],
        dataset_name: Optional[str],
        chat_id: str,
        chat_name: str,
        aggregation_value: Union[float, List[Dict[str, float]]],
        aggregation_type: AggregationType,
        creation_date: str,
        creation_user_id: str,
        tenant_id: str,
    ) -> None:
        self.llm_blueprint_id = llm_blueprint_id
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.ootb_dataset_name = ootb_dataset_name
        self.metric_name = metric_name
        self.deployment_id = deployment_id
        self.dataset_id = dataset_id
        self.dataset_name = dataset_name
        self.chat_id = chat_id
        self.chat_name = chat_name
        self.aggregation_value = aggregation_value
        self.aggregation_type = aggregation_type
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.tenant_id = tenant_id

    @classmethod
    def create(
        cls,
        chat_name: str,
        llm_blueprint_ids: List[str],
        evaluation_dataset_configuration_id: str,
        insights_configuration: list[InsightsConfiguration],
    ) -> Any:
        """
        Create a new evaluation dataset metric aggregation job.  The job will run the
        specified metric for the specified LLM blueprint IDs using the prompt-response pairs in
        the evaluation dataset.

        Parameters
        ----------
        chat_name : str
            The name of the chat that will be created to run the evaluation in.
        llm_blueprint_ids : List[str]
            The LLM blueprint IDs to evaluate.
        evaluation_dataset_configuration_id : str
            The ID evaluation dataset configuration to use during the evaluation.
        insights_configuration : List[InsightsConfiguration]
            The insights configurations to use during the evaluation.

        Returns
        -------
        str
            The ID of the evaluation dataset metric aggregation job.
        """
        payload = {
            "chat_name": chat_name,
            "llm_blueprint_ids": llm_blueprint_ids,
            "evaluation_dataset_configuration_id": evaluation_dataset_configuration_id,
            "insights_configuration": [insight.to_dict() for insight in insights_configuration],
        }
        url = f"{cls._client.domain}/{cls._path}/"
        response_data = cls._client.post(url, json=payload)
        return response_data.json()["jobId"]

    @classmethod
    def list(
        cls,
        llm_blueprint_ids: Optional[List[str]] = None,
        chat_ids: Optional[List[str]] = None,
        evaluation_dataset_configuration_ids: Optional[List[str]] = None,
        metric_names: Optional[List[str]] = None,
        aggregation_types: Optional[List[str]] = None,
        current_configuration_only: Optional[bool] = False,
        sort: Optional[str] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 100,
        non_errored_only: Optional[bool] = True,
    ) -> List[EvaluationDatasetMetricAggregation]:
        """List evaluation dataset metric aggregations.  The results will be filtered by the provided
        LLM blueprint IDs and chat IDs.

        Parameters
        ----------
        llm_blueprint_ids : List[str]
            The LLM blueprint IDs to filter on.
        chat_ids : List[str]
            The chat IDs to filter on.
        evaluation_dataset_configuration_ids : List[str]
            The evaluation dataset configuration IDs to filter on.
        metric_names : List[str]
            The metric names to filter on.
        aggregation_types : List[str]
            The aggregation types to filter on.
        current_configuration_only : Optional[bool]
            If True, only results that are associated with the current configuration of the LLM blueprint
            will be returned.  Defaults to False.
        sort : Optional[str]
            The field to sort on.  Defaults to None.
        offset : Optional[int]
            The offset to start at.  Defaults to 0.
        limit : Optional[int]
            The maximum number of results to return.  Defaults to 100.
        non_errored_only : Optional[bool]
            If True, only results that did not encounter an error will be returned.  Defaults to True.

        Returns
        -------
        List[EvaluationDatasetMetricAggregation]
            A list of evaluation dataset metric aggregations.
        """
        params = {
            "llm_blueprint_ids": llm_blueprint_ids,
            "chat_ids": chat_ids,
            "evaluation_dataset_configuration_ids": evaluation_dataset_configuration_ids,
            "metric_names": metric_names,
            "aggregation_types": aggregation_types,
            "current_configuration_only": current_configuration_only,
            "sort": sort,
            "offset": offset,
            "limit": limit,
            "non_errored_only": non_errored_only,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    @classmethod
    def delete(cls, llm_blueprint_ids: Optional[List[str]], chat_ids: Optional[List[str]]) -> None:
        """Delete the associated evaluation dataset metric aggregations.  Either llm_blueprint_ids
        or chat_ids must be provided.  If both are provided, only results matching both will be removed.

        Parameters
        ----------
        llm_blueprint_ids : List[str]
            The LLM blueprint IDs to filter on.
        chat_ids : List[str]
            The chat IDs to filter on.

        """
        if llm_blueprint_ids is None and chat_ids is None:
            raise ValueError("Either llm_blueprint_ids or chat_ids must be provided.")
        url = f"{cls._client.domain}/{cls._path}/"
        # parameters are camelcase here because delete doesn't seem to send them correctly otherwise.
        params = {
            "llmBlueprintIds": llm_blueprint_ids,
            "chatIds": chat_ids,
        }
        cls._client.delete(url, params=params)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.genai.chat_prompt import confidence_scores_trafaret, result_metadata_trafaret
from datarobot.models.genai.comparison_chat import get_entity_id
from datarobot.models.genai.llm_blueprint import vector_database_settings_trafaret
from datarobot.models.genai.playground import Playground
from datarobot.utils.pagination import unpaginate

trace_chat = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
    }
).ignore_extra("*")

datarobot_user = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String(allow_blank=True),
    }
).ignore_extra("*")

trace_metadata = t.Dict(
    {
        t.Key("users"): t.List(datarobot_user),
        t.Key("chats"): t.List(trace_chat),
    }
).ignore_extra("*")


prompt_trace = t.Dict(
    {
        t.Key("timestamp"): t.String,
        t.Key("user"): datarobot_user,
        t.Key("chat_prompt_id", optional=True): t.Or(t.String, t.Null),
        t.Key("use_case_id", optional=True): t.Or(t.String, t.Null),
        t.Key("comparison_prompt_id", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_blueprint_id"): t.String,
        t.Key("llm_blueprint_name"): t.String,
        t.Key("llm_name", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_vendor", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_license", optional=True): t.Or(t.String, t.Null),
        t.Key("llm_settings", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        t.Key("chat_name", optional=True): t.Or(t.String, t.Null),
        t.Key("chat_id", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_id", optional=True): t.Or(t.String, t.Null),
        t.Key("vector_database_settings", optional=True): t.Or(
            vector_database_settings_trafaret, t.Null
        ),
        t.Key("result_metadata", optional=True): t.Or(
            result_metadata_trafaret, t.Dict().allow_extra("*")
        ),
        t.Key("result_text", optional=True): t.Or(t.String, t.Null),
        t.Key("confidence_scores", optional=True): t.Or(confidence_scores_trafaret, t.Null),
        t.Key("text"): t.String,
        t.Key("execution_status"): t.String,
        t.Key("prompt_type", optional=True): t.Or(t.String, t.Null),
        t.Key("evaluation_dataset_configuration_id", optional=True): t.Or(t.String, t.Null),
        t.Key("warning", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


class PromptTrace(APIObject):
    """
    Prompt trace contains aggregated information about a prompt execution.

    Attributes
    ----------
    timestamp : str
        The timestamp of the trace (ISO 8601 formatted).
    user : dict
        The user who submitted the prompt.
    chat_prompt_id : str
        The ID of the chat prompt associated with the trace.
    use_case_id : str
        The ID of the Use Case the playground is in.
    comparison_prompt_id : str
        The ID of the comparison prompts associated with the trace.
    llm_blueprint_id : str
        The ID of the LLM blueprint that the prompt was submitted to.
    llm_blueprint_name: str
        The name of the LLM blueprint.
    llm_name : str
        The name of the LLM in the LLM blueprint.
    llm_vendor : str
        The vendor name of the LLM.
    llm_license : str
        What type of license the LLM has.
    llm_settings : dict or None
        The LLM settings for the LLM blueprint. The specific keys allowed and the
        constraints on the values are defined in the response from `LLMDefinition.list`,
        but this typically has dict fields. Either:
        - system_prompt - The system prompt that influences the LLM responses.
        - max_completion_length - The maximum number of tokens in the completion.
        - temperature - Controls the variability in the LLM response.
        - top_p - Sets whether the model considers next tokens with top_p probability mass.
        or
        - system_prompt - The system prompt that influences the LLM responses.
        - validation_id - The ID of the external model LLM validation.
        - external_llm_context_size - The external LLM's context size, in tokens,
        for external model-based LLM blueprints.
    chat_name: str or None
        The name of the chat associated with the Trace.
    chat_id: str or None
        The ID of the chat associated with the Trace.
    vector_database_id : str or None
        ID of the vector database associated with the LLM blueprint, if any.
    vector_database_settings : VectorDatabaseSettings or None
        The settings for the vector database associated with the LLM blueprint, if any.
    result_metadata : ResultMetadata or None
        Metadata for the result of the prompt submission.
    result_text: str or None
        The result text from the prompt submission.
    confidence_scores: ConfidenceScores or None
        The confidence scores if there is a vector database associated with the prompt.
    text: str
        The prompt text submitted to the LLM.
    execution_status: str
        The execution status of the chat prompt.
    prompt_type: str or None
        The type of prompting strategy, for example history aware.
    evaluation_dataset_configuration_id: str or None
        The ID of the evaluation dataset configuration associated with the trace.
    warning: str or None
        Any warnings associated with the trace.
    """

    _path = "api/v2/genai/playgrounds/{}/trace/"
    _export_path = "api/v2/genai/playgrounds/{}/traceDatasets/"

    _converter = prompt_trace

    def __init__(
        self,
        timestamp: str,
        user: Dict[str, str],
        use_case_id: str,
        llm_blueprint_id: str,
        llm_blueprint_name: str,
        text: str,
        execution_status: str,
        llm_name: Optional[str] = None,
        llm_vendor: Optional[str] = None,
        llm_license: Optional[str] = None,
        chat_prompt_id: Optional[str] = None,
        comparison_prompt_id: Optional[str] = None,
        llm_settings: Optional[Dict[str, Any]] = None,
        chat_name: Optional[str] = None,
        chat_id: Optional[str] = None,
        vector_database_id: Optional[str] = None,
        vector_database_settings: Optional[Dict[str, Any]] = None,
        result_metadata: Optional[Dict[str, Any]] = None,
        result_text: Optional[str] = None,
        confidence_scores: Optional[Dict[str, float]] = None,
        prompt_type: Optional[str] = None,
        evaluation_dataset_configuration_id: Optional[str] = None,
        warning: Optional[str] = None,
    ):
        self.timestamp = timestamp
        self.user = user
        self.chat_prompt_id = (chat_prompt_id,)
        self.use_case_id = use_case_id
        self.comparison_prompt_id = comparison_prompt_id
        self.llm_blueprint_id = llm_blueprint_id
        self.llm_blueprint_name = llm_blueprint_name
        self.llm_name = llm_name
        self.llm_vendor = llm_vendor
        self.llm_license = llm_license
        self.llm_settings = llm_settings
        self.chat_name = chat_name
        self.chat_id = chat_id
        self.vector_database_id = vector_database_id
        self.vector_database_settings = vector_database_settings
        self.result_metadata = result_metadata
        self.result_text = result_text
        self.confidence_scores = confidence_scores
        self.text = text
        self.execution_status = execution_status
        self.prompt_type = prompt_type
        self.evaluation_dataset_configuration_id = evaluation_dataset_configuration_id
        self.warning = warning

    @classmethod
    def list(cls, playground: Union[str, Playground]) -> List[PromptTrace]:
        """
        List all prompt traces for a playground.

        Parameters
        ----------
        playground: str
            The ID of the playground to list prompt traces for.

        Returns
        -------
        prompt_traces: list[PromptTrace]
            List of prompt traces for the playground.
        """
        playground_id = get_entity_id(playground)
        url = f"{cls._client.domain}/{cls._path.format(playground_id)}"
        response = unpaginate(url, None, cls._client)
        return [cls.from_server_data(trace) for trace in response]

    @classmethod
    def export_to_ai_catalog(cls, playground: Union[str, Playground]) -> Any:
        """
        Export prompt traces to AI Catalog as a CSV.

        Parameters
        ----------
        playground: str
            The ID of the playground to export prompt traces for.

        Returns
        -------
        status_url: str
            The URL where the status of the job can be monitored
        """
        playground_id = get_entity_id(playground)
        url = f"{cls._client.domain}/{cls._export_path.format(playground_id)}"
        response = cls._client.post(url)
        return response.headers["Location"]


class TraceMetadata(APIObject):
    """
    Trace metadata contains information about all the users and chats that are relevant to
    this playground.

    Attributes
    ----------
    users : list[dict]
        The users who submitted the prompt.
    """

    _path = "api/v2/genai/playgrounds/{}/trace/metadata/"

    _converter = trace_metadata

    def __init__(self, users: List[Dict[str, str]], chats: List[Dict[str, str]]):
        self.users = users
        self.chats = chats

    @classmethod
    def get(cls, playground: Union[str, Playground]) -> TraceMetadata:
        """
        Get trace metadata for a playground.

        Parameters
        ----------
        playground: str
            The ID of the playground to get trace metadata for.

        Returns
        -------
        trace_metadata: TraceMetadata
            The trace metadata for the playground.
        """
        playground_id = get_entity_id(playground)
        url = f"{cls._client.domain}/{cls._path.format(playground_id)}"
        response = cls._client.get(url)
        return cls.from_server_data(response.json())



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, cast, Dict, List, Optional, Type, Union
from urllib.parse import urlencode

import trafaret as t

from datarobot.enums import (
    DataWranglingDataSourceTypes,
    DataWranglingDialect,
    DataWranglingSnapshotPolicy,
    DEFAULT_MAX_WAIT,
    enum_to_list,
    RecipeInputType,
    RecipeType,
)
from datarobot.models.api_object import APIObject
from datarobot.models.data_store import DataStore
from datarobot.models.dataset import Dataset
from datarobot.models.recipe_operation import (
    DatetimeSamplingOperation,
    DownsamplingOperation,
    RandomSamplingOperation,
    SamplingOperation,
    WranglingOperation,
)
from datarobot.models.use_cases.use_case import UseCase
from datarobot.utils import to_api
from datarobot.utils.waiters import wait_for_async_resolution


class DataSourceInput(APIObject):
    """Inputs required to create a new recipe from data store."""

    _converter = t.Dict(
        {
            t.Key("canonical_name"): t.String,
            t.Key("table"): t.String,
            t.Key("schema", optional=True): t.Or(t.String(), t.Null),
            t.Key("catalog", optional=True): t.Or(t.String(), t.Null),
            t.Key("sampling", optional=True): t.Or(SamplingOperation._converter, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        canonical_name: str,
        table: str,
        schema: Optional[str] = None,
        catalog: Optional[str] = None,
        sampling: Optional[Union[RandomSamplingOperation, DatetimeSamplingOperation]] = None,
    ):
        self.canonical_name = canonical_name
        self.table = table
        self.schema = schema
        self.catalog = catalog
        self.sampling = sampling


class DatasetInput(APIObject):
    _converter = t.Dict(
        {
            t.Key("sampling"): SamplingOperation._converter,
        }
    ).allow_extra("*")

    def __init__(self, sampling: SamplingOperation):
        self.sampling = (
            SamplingOperation.from_server_data(sampling) if isinstance(sampling, dict) else sampling
        )


class RecipeDatasetInput(APIObject):
    """Object, describing inputs for recipe transformations."""

    _converter = t.Dict(
        {
            t.Key("input_type"): t.Atom(RecipeInputType.DATASET),
            t.Key("dataset_id"): t.String,
            t.Key("dataset_version_id", optional=True): t.Or(t.String, t.Null),
            t.Key("snapshot_policy", optional=True): t.Or(
                t.Enum(*enum_to_list(DataWranglingSnapshotPolicy)), t.Null
            ),
            t.Key("sampling", optional=True): t.Or(SamplingOperation._converter, t.Null),
            t.Key("alias", optional=True): t.Or(t.String(), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        input_type: RecipeInputType,
        dataset_id: str,
        dataset_version_id: Optional[str] = None,
        snapshot_policy: Optional[DataWranglingSnapshotPolicy] = DataWranglingSnapshotPolicy.LATEST,
        sampling: Optional[Union[SamplingOperation, Dict[str, Any]]] = None,
        alias: Optional[str] = None,
    ):
        self.input_type = input_type
        self.dataset_id = dataset_id
        self.snapshot_policy = snapshot_policy
        self.dataset_version_id = (
            dataset_version_id if snapshot_policy != DataWranglingSnapshotPolicy.LATEST else None
        )
        self.sampling = (
            SamplingOperation.from_server_data(sampling) if isinstance(sampling, dict) else sampling
        )
        self.alias = alias

    def to_api(
        self, keep_attrs: Optional[Union[List[str], List[List[str]]]] = None
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        res = cast(Dict[str, Any], to_api(self, keep_attrs=keep_attrs))
        if self.snapshot_policy == DataWranglingSnapshotPolicy.LATEST:
            res["dataset_version_id"] = None
        return res


class JDBCTableDataSourceInput(APIObject):
    """Object, describing inputs for recipe transformations."""

    _converter = t.Dict(
        {
            t.Key("input_type"): t.Atom(RecipeInputType.DATASOURCE),
            t.Key("data_source_id"): t.String,
            t.Key("data_store_id"): t.String,
            t.Key("dataset_id", optional=True): t.Or(t.String(), t.Null),
            t.Key("sampling", optional=True): t.Or(SamplingOperation._converter, t.Null),
            t.Key("alias", optional=True): t.Or(t.String(), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        input_type: RecipeInputType,
        data_source_id: str,
        data_store_id: str,
        dataset_id: Optional[str] = None,
        sampling: Optional[Union[SamplingOperation, Dict[str, Any]]] = None,
        alias: Optional[str] = None,
    ):
        self.input_type = input_type
        self.data_source_id = data_source_id
        self.data_store_id = data_store_id
        self.dataset_id = dataset_id
        self.sampling = (
            SamplingOperation.from_server_data(sampling) if isinstance(sampling, dict) else sampling
        )
        self.alias = alias

    def to_api(
        self, keep_attrs: Optional[Union[List[str], List[List[str]]]] = None
    ) -> Union[str, Dict[str, Any], List[Dict[str, Any]]]:
        return to_api(self, keep_attrs=keep_attrs)


class RecipeSettings(APIObject):
    """Settings, for example to apply at downsampling stage."""

    _converter = t.Dict(
        {
            t.Key("target", optional=True): t.Or(t.String(), t.Null),
            t.Key("weights_feature", optional=True): t.Or(t.String(), t.Null),
            t.Key("prediction_point", optional=True, default=None): t.Or(t.String(), t.Null),
            t.Key("relationships_configuration_id", optional=True, default=None): t.Or(
                t.String(), t.Null
            ),
            t.Key(
                "feature_discovery_supervised_feature_reduction", optional=True, default=None
            ): t.Or(t.Bool(), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        target: Optional[str] = None,
        weights_feature: Optional[str] = None,
        prediction_point: Optional[str] = None,
        relationships_configuration_id: Optional[str] = None,
        feature_discovery_supervised_feature_reduction: Optional[bool] = None,
    ):
        self.target = target
        self.weights_feature = weights_feature
        self.prediction_point = prediction_point
        self.relationships_configuration_id = relationships_configuration_id
        self.feature_discovery_supervised_feature_reduction = (
            feature_discovery_supervised_feature_reduction
        )


class RecipeMetadata(APIObject):
    """The recipe metadata includes the name, description, recipe type, and sql"""

    _converter = t.Dict(
        {
            t.Key("name", optional=True): t.String(),
            t.Key("description", optional=True): t.String(allow_blank=True, max_length=1000),
            t.Key("recipe_type", optional=True): t.Enum(*enum_to_list(RecipeType)),
            t.Key("sql", optional=True): t.String(allow_blank=True, max_length=64000),
        }
    )

    def __init__(
        self,
        name: Optional[str],
        description: Optional[str],
        recipe_type: Optional[RecipeType],
        sql: Optional[str],
    ):
        self.name = name
        self.description = description
        self.recipe_type = recipe_type
        self.sql = sql


class Recipe(APIObject):
    """Data wrangling entity, which contains all information needed to transform dataset and generate SQL."""

    _path = "recipes/"

    _converter = t.Dict(
        {
            t.Key("dialect"): t.String,
            t.Key("recipe_id"): t.String,
            t.Key("status"): t.String,
            t.Key("inputs"): t.List(
                t.Or(JDBCTableDataSourceInput._converter, RecipeDatasetInput._converter)
            ),
            t.Key("operations", optional=True): t.List(WranglingOperation._converter),
            t.Key("downsampling", optional=True): t.Or(DownsamplingOperation._converter, t.Null),
            t.Key("settings", optional=True): t.Or(RecipeSettings._converter, t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        dialect: DataWranglingDialect,
        recipe_id: str,
        status: str,
        inputs: List[Dict[str, Any]],
        operations: Optional[List[Dict[str, Any]]] = None,
        downsampling: Optional[Dict[str, Any]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.dialect = dialect
        self.id = recipe_id
        self.status = status
        self.inputs = []
        for input in inputs:
            input_clz: Union[Type[JDBCTableDataSourceInput], Type[RecipeDatasetInput]]
            if input["input_type"] == RecipeInputType.DATASOURCE:
                input_clz = JDBCTableDataSourceInput
            elif input["input_type"] == RecipeInputType.DATASET:
                input_clz = RecipeDatasetInput
            else:
                raise RuntimeError("unknown input_type")
            self.inputs.append(input_clz.from_server_data(input))
        self.operations = (
            [WranglingOperation.from_server_data(op) for op in operations]
            if operations is not None
            else None
        )
        self.downsampling = (
            DownsamplingOperation.from_server_data(downsampling)
            if isinstance(downsampling, dict)
            else downsampling
        )
        self.settings = (
            RecipeSettings.from_server_data(settings) if isinstance(settings, dict) else settings
        )

    @classmethod
    def update_downsampling(cls, recipe_id: str, downsampling: DownsamplingOperation) -> Recipe:
        """Set downsampling for the recipe, applied during publishing."""
        path = f"{cls._path}{recipe_id}/downsampling/"
        payload = {"downsampling": to_api(downsampling)}
        response = cls._client.put(path, json=payload)
        return Recipe.from_server_data(response.json())

    def retrieve_preview(
        self, max_wait: int = DEFAULT_MAX_WAIT, number_of_operations_to_use: Optional[int] = None
    ) -> Dict[str, Any]:
        """Retrieve preview and compute it, if absent.

        Parameters
        ----------
        max_wait: int
            The number of seconds to wait for the result.
        number_of_operations_to_use: Optional[int]
            Request preview for particular number of operations.

        Returns
        -------
        preview: dict

        """
        path = f"{self._path}{self.id}/preview/"
        payload = {}
        if number_of_operations_to_use is not None:
            payload = {"numberOfOperationsToUse": number_of_operations_to_use}
        response = self._client.post(path, json=payload)
        finished_url = wait_for_async_resolution(
            self._client, response.headers["Location"], max_wait=max_wait
        )
        r_data = self._client.get(finished_url).json()
        # TODO: create an ApiObject for Preview
        return r_data  # type: ignore[no-any-return]

    def retrieve_insights(
        self, max_wait: int = DEFAULT_MAX_WAIT, number_of_operations_to_use: Optional[int] = None
    ) -> Any:
        """Retrieve insights for the sample. When preview is requested, the insights job starts automatically.

        Parameters
        ----------
        max_wait: int
            The number of seconds to wait for the result.
        number_of_operations_to_use: Optional[int]
            Retrieves insights for the specified number of operations. First, preview computation for the same
            number of operations must be submitted.

        Returns
        -------

        """
        url = f"recipes/{self.id}/insights/"
        if number_of_operations_to_use is not None:
            query_params = {"numberOfOperationsToUse": number_of_operations_to_use}
            url = f"{url}?{urlencode(query_params)}"
        return wait_for_async_resolution(self._client, url, max_wait)

    @classmethod
    def set_inputs(
        cls, recipe_id: str, inputs: List[Union[JDBCTableDataSourceInput, RecipeDatasetInput]]
    ) -> Recipe:
        """Set inputs for the recipe."""
        path = f"{cls._path}{recipe_id}/inputs/"
        payload = {"inputs": [input_.to_api() for input_ in inputs]}
        response = cls._client.put(path, json=payload)
        return Recipe.from_server_data(response.json())

    @classmethod
    def set_operations(cls, recipe_id: str, operations: List[WranglingOperation]) -> Recipe:
        """Set operations for the recipe."""
        path = f"{cls._path}{recipe_id}/operations/"
        payload = {"operations": [to_api(operation) for operation in operations]}
        response = cls._client.put(path, json=payload)
        return Recipe.from_server_data(response.json())

    @classmethod
    def set_recipe_metadata(cls, recipe_id: str, metadata: Dict[str, str]) -> Recipe:
        """
        Update metadata for the recipe.

        Parameters
        ----------
        recipe_id: str
            Recipe ID.
        metadata: Dict[str, str]
            Dictionary of metadata to be updated.

        Returns
        -------
        recipe: Recipe
            New recipe with updated metadata.
        """
        RecipeMetadata._converter.check(metadata)
        path = f"{cls._path}{recipe_id}/"
        response = cls._client.patch(path, json=metadata)
        return Recipe.from_server_data(response.json())

    @classmethod
    def get(cls, recipe_id: str) -> Recipe:
        path = f"{cls._path}{recipe_id}/"
        return cls.from_location(path)

    def get_sql(self, operations: Optional[List[WranglingOperation]] = None) -> str:
        """Generate sql for the given recipe in a transient way, recipe is not modified.
        if operations is None, recipe operations are used to generate sql.
        if operations = [], recipe operations are ignored during sql generation.
        if operations is not empty list, generate sql for them.
        """
        path = f"{self._path}{self.id}/sql/"
        payload = {
            "operations": (
                [to_api(operation) for operation in operations] if operations else operations
            )
        }
        response = self._client.post(path, data=payload)
        return response.json()["sql"]  # type: ignore[no-any-return]

    @classmethod
    def from_data_store(
        cls,
        use_case: UseCase,
        data_store: DataStore,
        data_source_type: DataWranglingDataSourceTypes,
        dialect: DataWranglingDialect,
        data_source_inputs: List[DataSourceInput],
        recipe_type: Optional[RecipeType] = RecipeType.WRANGLING,
    ) -> Recipe:
        """Create a wrangling recipe from data store."""
        payload = {
            "use_case_id": use_case.id,
            "data_store_id": data_store.id,
            "data_source_type": data_source_type,
            "dialect": dialect,
            "inputs": [to_api(input_) for input_ in data_source_inputs],
            "recipe_type": recipe_type,
        }
        path = f"{cls._path}fromDataStore/"
        response = cls._client.post(path, data=payload)
        return Recipe.from_server_data(response.json())

    @classmethod
    def from_dataset(
        cls,
        use_case: UseCase,
        dataset: Dataset,
        dialect: Optional[DataWranglingDialect] = None,
        inputs: Optional[List[DatasetInput]] = None,
        recipe_type: Optional[RecipeType] = RecipeType.WRANGLING,
        snapshot_policy: Optional[DataWranglingSnapshotPolicy] = DataWranglingSnapshotPolicy.LATEST,
    ) -> Recipe:
        """Create a wrangling recipe from dataset."""

        payload = {
            "use_case_id": use_case.id,
            "dataset_id": dataset.id,
            "dataset_version_id": (
                None
                if snapshot_policy == DataWranglingSnapshotPolicy.LATEST
                else dataset.version_id
            ),
            "dialect": dialect,
            "inputs": [to_api(input) for input in inputs] if inputs else None,
            "recipe_type": recipe_type,
            "snapshot_policy": snapshot_policy,
        }
        path = f"{cls._path}fromDataset/"
        response = cls._client.post(path, data=payload)
        return Recipe.from_server_data(response.json())



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import date
from typing import Any, Dict, List, Optional, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.use_cases.use_case import UseCase
from datarobot.models.use_cases.utils import get_use_case_id
from datarobot.utils import underscorize
from datarobot.utils.pagination import unpaginate


class BooleanConstraintsDict(TypedDict):
    type: str


class NumericConstraintsDict(TypedDict):
    type: str
    min_value: Optional[Union[int, float]]
    max_value: Optional[Union[int, float]]


class StringConstraintsDict(TypedDict):
    type: str
    min_length: Optional[int]
    max_length: Optional[int]
    allowed_choices: Optional[List[str]]


class GeneralConstraintsDict(TypedDict):
    type: str
    min_value: Optional[Union[int, float]]
    max_value: Optional[Union[int, float]]
    min_length: Optional[int]
    max_length: Optional[int]
    allowed_choices: Optional[list[str]]


class LLMSettingDefinitionDict(TypedDict):
    """Dict representation of LLMSettingDefinition."""

    id: str
    name: str
    description: str
    type: str
    format: Optional[str]
    is_nullable: bool
    constraints: Optional[
        Union[BooleanConstraintsDict, NumericConstraintsDict, StringConstraintsDict]
    ]
    default_value: Optional[Union[bool, int, float, str]]


class LLMDefinitionDict(TypedDict):
    """Dict representation of LLMDefinition."""

    id: str
    name: str
    description: str
    vendor: str
    license: str
    supported_languages: str
    settings: List[LLMSettingDefinitionDict]
    context_size: Optional[int]
    retirement_date: Optional[date]
    is_deprecated: Optional[bool]
    is_active: Optional[bool]


llm_setting_constraints_trafaret = t.Dict(
    {
        t.Key("type"): t.String,
        t.Key("min_value", optional=True): t.Or(t.Int, t.Float, t.Null),
        t.Key("max_value", optional=True): t.Or(t.Int, t.Float, t.Null),
        t.Key("min_length", optional=True): t.Or(t.Int, t.Null),
        t.Key("max_length", optional=True): t.Or(t.Int, t.Null),
        t.Key("allowed_choices", optional=True): t.Or(t.List[t.String], t.Null),
    }
).ignore_extra("*")

llm_setting_definition_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("description"): t.String,
        t.Key("type"): t.String,
        t.Key("format", optional=True): t.Or(t.String, t.Null),
        t.Key("is_nullable"): t.Bool,
        t.Key("constraints"): t.Or(llm_setting_constraints_trafaret, t.Null),
        t.Key("default_value", optional=True): t.Or(
            t.Null, t.Bool, t.Int, t.Float, t.String(allow_blank=True)
        ),
    }
).ignore_extra("*")

language_model_definition_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("name"): t.String,
        t.Key("description"): t.String,
        t.Key("vendor"): t.String,
        t.Key("license"): t.String,
        t.Key("supported_languages"): t.String,
        t.Key("settings"): t.List(llm_setting_definition_trafaret),
        t.Key("context_size", optional=True): t.Or(t.Int, t.Null),
        t.Key("is_deprecated", optional=True): t.Bool,
        t.Key("is_active", optional=True): t.Bool,
        t.Key("retirement_date", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


class LLMSettingConstraint(APIObject):
    """
    Metadata for DataRobot GenAI LLMSettingConstraint.
    Attributes
    ----------
    type : str
        Data type of the setting.
    min_value : int, float, or None, optional
        Minimum value for the setting.
    max_value : int, float, or None, optional
        Maximum value for the setting.
    min_length : int or None, optional
        The minimum length for a setting of string type.
    max_length : int or None, optional
        The maximum length for a setting of string type.
    allowed_choices : list[str] or None, optional
        Allowed values for a setting of string type.
    """

    _converter = llm_setting_constraints_trafaret

    def __init__(
        self,
        type: str,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        allowed_choices: Optional[List[str]] = None,
    ):
        self.type = type
        self.min_value = min_value
        self.max_value = max_value
        self.min_length = min_length
        self.max_length = max_length
        self.allowed_choices = allowed_choices

    def to_dict(
        self,
    ) -> Union[
        BooleanConstraintsDict,
        NumericConstraintsDict,
        StringConstraintsDict,
        GeneralConstraintsDict,
    ]:
        """Convert LLMSettingContraint to dict representation."""
        if self.type == "boolean":
            return {"type": self.type}
        if self.type in {"integer", "float"}:
            return {
                "type": self.type,
                "min_value": self.min_value,
                "max_value": self.max_value,
            }
        if self.type == "string":
            return {
                "type": self.type,
                "min_length": self.min_length,
                "max_length": self.max_length,
                "allowed_choices": self.allowed_choices,
            }

        return {
            "type": self.type,
            "min_value": self.min_value,
            "max_value": self.max_value,
            "min_length": self.min_length,
            "max_length": self.max_length,
            "allowed_choices": self.allowed_choices,
        }


class LLMSettingDefinition(APIObject):
    """
    Metadata for DataRobot GenAI LLMSetting.
    Attributes
    ----------
    id : str
        The seetting ID.
    name : str
        The setting name.
    description : str
        Description of the setting.
    type : str
        The data type associated with this setting.
    format : str or None
        The string format, multiline or None.
    is_nullable : bool
        Whether the setting is nullable.
    constraints : LLMSettingConstraint or None
        Constraints for the setting.
    default_value : int, float, string, bool, or None
        The default value for the setting.
    """

    _converter = llm_setting_definition_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        type: str,
        is_nullable: bool,
        format: Optional[str] = None,
        constraints: Optional[Dict[str, Any]] = None,
        default_value: Optional[Union[int, float, str, bool]] = None,
    ):
        self.id = underscorize(id)
        self.name = name
        self.description = description
        self.type = type
        self.format = format
        self.is_nullable = is_nullable
        self.constraints = (
            LLMSettingConstraint.from_server_data(constraints) if constraints else None
        )
        self.default_value = default_value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name})"

    def to_dict(self) -> LLMSettingDefinitionDict:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "format": self.format,
            "is_nullable": self.is_nullable,
            "constraints": self.constraints.to_dict() if self.constraints else None,
            "default_value": self.default_value,
        }


class LLMDefinition(APIObject):
    """
    Metadata for a DataRobot GenAI LLM.

    Attributes
    ----------
    id : str
        Language model type ID.
    name : str
        Language model name.
    description : str
        Description of the language model.
    vendor : str
        Name of the vendor for this model.
    license : str
        License for this model.
    supported_languages : str
        Languages supported by this model.
    settings : list of LLMSettingDefinition
        Settings for this model
    context_size : int
        The context size for this model
    retirement_date: datetime.date, optional
        When the LLM is expected to be retired and no longer
        available for submitting new prompts.
    is_deprecated : bool, optional
        Whether the LLM is deprecated and will be removed when
        it is past the retirement date.
    is_active : bool, optional
        Whether the LLM is active.
    """

    _path = "api/v2/genai/llms"

    _converter = language_model_definition_trafaret

    def __init__(
        self,
        id: str,
        name: str,
        description: str,
        vendor: str,
        license: str,
        supported_languages: str,
        settings: List[Dict[str, Any]],
        context_size: Optional[int] = None,
        retirement_date: Optional[date] = None,
        is_deprecated: Optional[bool] = False,
        is_active: Optional[bool] = True,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.vendor = vendor
        self.license = license
        self.supported_languages = supported_languages
        self.settings = [LLMSettingDefinition.from_server_data(setting) for setting in settings]
        self.context_size = context_size
        self.retirement_date = retirement_date
        self.is_deprecated = is_deprecated
        self.is_active = is_active

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name={self.name})"

    @classmethod
    def list(
        cls, use_case: Optional[Union[UseCase, str]] = None, as_dict: bool = True
    ) -> Union[List[LLMDefinition], List[LLMDefinitionDict]]:
        """
        List all large language models (LLMs) available to the user.

        Parameters
        ----------
        use_case : Optional[UseCase or str], optional
            The returned LLMs, including external LLMs, available
            for the specified Use Case.
            Accepts either the entity or the Use CaseID.

        Returns
        -------
        llms : list[LLMDefinition] or list[LLMDefinitionDict]
            A list of large language models (LLMs) available to the user.
        """
        url = f"{cls._client.domain}/{cls._path}/"
        params = {"use_case_id": get_use_case_id(use_case, is_required=False)}
        r_data = unpaginate(url, params, cls._client)
        llms = [cls.from_server_data(data) for data in r_data]

        if as_dict:
            return [llm.to_dict() for llm in llms]
        return llms

    def to_dict(self) -> LLMDefinitionDict:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "vendor": self.vendor,
            "license": self.license,
            "supported_languages": self.supported_languages,
            "settings": [setting.to_dict() for setting in self.settings],
            "context_size": self.context_size,
            "retirement_date": self.retirement_date,
            "is_deprecated": self.is_deprecated,
            "is_active": self.is_active,
        }



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import contextlib
import json
import os
from typing import List, Optional

from requests_toolbelt import MultipartEncoder
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import CustomTaskOutboundNetworkPolicy
from datarobot.models.api_object import APIObject
from datarobot.models.custom_model_version import (
    CustomDependency,
    CustomModelFileItem,
    RequiredMetadataValue,
)
from datarobot.models.custom_task_version_dependency_build import CustomTaskVersionDependencyBuild
from datarobot.models.trafarets import UserBlueprintTaskArgument_
from datarobot.models.user_blueprints.models import UserBlueprintTaskArgument
from datarobot.utils import camelize
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_custom_resolution


class CustomTaskFileItem(CustomModelFileItem):
    """A file item attached to a DataRobot custom task version.

    .. versionadded:: v2.26

    Attributes
    ----------
    id: str
        id of the file item
    file_name: str
        name of the file item
    file_path: str
        path of the file item
    file_source: str
        source of the file item
    created_at: str
        ISO-8601 formatted timestamp of when the version was created
    """

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("file_name"): String(),
            t.Key("file_path"): String(),
            t.Key("file_source"): String(),
            t.Key("created") >> "created_at": String(),
        }
    ).ignore_extra("*")

    schema = _converter


class CustomTaskVersion(APIObject):
    """A version of a DataRobot custom task.

    .. versionadded:: v2.26

    Attributes
    ----------
    id: str
        id of the custom task version
    custom_task_id: str
        id of the custom task
    version_minor: int
        a minor version number of custom task version
    version_major: int
        a major version number of custom task version
    label: str
        short human readable string to label the version
    created_at: str
        ISO-8601 formatted timestamp of when the version was created
    is_frozen: bool
        a flag if the custom task version is frozen
    items: List[CustomTaskFileItem]
        a list of file items attached to the custom task version
    description: Optional[str]
        custom task version description
    base_environment_id: Optional[str]
        id of the environment to use with the task
    base_environment_version_id: Optional[str]
        id of the environment version to use with the task
    dependencies: List[CustomDependency]
        the parsed dependencies of the custom task version if the
        version has a valid requirements.txt file
    required_metadata_values: List[RequiredMetadataValue]
        Additional parameters required by the execution environment. The required keys are
        defined by the fieldNames in the base environment's requiredMetadataKeys.
    arguments: List[UserBlueprintTaskArgument]
        A list of custom task version arguments.
    outbound_network_policy: CustomTaskOutboundNetworkPolicy
    """

    _path = "customTasks/{}/versions/"
    _dependency_build_path = "customTasks/{}/versions/{}/dependencyBuild/"
    _dependency_build_log_path = "customTasks/{}/versions/{}/dependencyBuildLog/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_task_id"): String(),
            t.Key("version_major"): Int(),
            t.Key("version_minor"): Int(),
            t.Key("label"): String(),
            t.Key("created") >> "created_at": String(),
            t.Key("is_frozen"): t.Bool(),
            t.Key("items"): t.List(CustomTaskFileItem.schema),
            # because `from_server_data` scrubs Nones, this must be optional here.
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("maximum_memory", optional=True): Int(),
            t.Key("base_environment_id", optional=True): t.Or(String(), t.Null()),
            t.Key("base_environment_version_id", optional=True): t.Or(String(), t.Null()),
            t.Key("dependencies", optional=True): t.List(CustomDependency.schema),
            t.Key("required_metadata_values", optional=True): t.List(RequiredMetadataValue.schema),
            t.Key("arguments", optional=True): t.List(UserBlueprintTaskArgument_),
            t.Key("outbound_network_policy"): t.Enum("ISOLATED", "PUBLIC"),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id,
        custom_task_id,
        version_major,
        version_minor,
        label,
        created_at,
        is_frozen,
        items,
        outbound_network_policy,
        description=None,
        base_environment_id=None,
        maximum_memory=None,
        base_environment_version_id=None,
        dependencies=None,
        required_metadata_values=None,
        arguments=None,
    ):
        if dependencies is None:
            dependencies = []
        arguments = arguments or []

        self.id = id
        self.custom_task_id = custom_task_id
        self.description = description
        self.version_major = version_major
        self.version_minor = version_minor
        self.label = label
        self.created_at = created_at
        self.is_frozen = is_frozen
        self.items = [CustomTaskFileItem(**data) for data in items]

        self.maximum_memory = maximum_memory
        self.required_metadata_values = (
            [RequiredMetadataValue(**val) for val in required_metadata_values]
            if required_metadata_values
            else None
        )
        self.base_environment_id = base_environment_id
        self.base_environment_version_id = base_environment_version_id
        self.dependencies = [CustomDependency(**data) for data in dependencies]
        self.arguments = [UserBlueprintTaskArgument(**argument) for argument in arguments]
        self.outbound_network_policy = CustomTaskOutboundNetworkPolicy.from_string(
            outbound_network_policy
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.label or self.id!r})"

    def _update_values(self, new_response: CustomTaskVersion) -> None:
        for attr in self._fields():
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def _all_versions_path(cls, task_id: str) -> str:
        return cls._path.format(task_id)

    @classmethod
    def _single_version_path(cls, task_id: str, version_id: str) -> str:
        return cls._all_versions_path(task_id) + f"{version_id}/"

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        initial = super().from_server_data(data, keep_attrs)
        # from_server_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        initial.required_metadata = data.get("requiredMetadata")
        return initial

    @classmethod
    def create_clean(
        cls,
        custom_task_id: str,
        base_environment_id: str,
        maximum_memory: Optional[int] = None,
        is_major_update: bool = True,
        folder_path: Optional[str] = None,
        required_metadata_values: Optional[List[RequiredMetadataValue]] = None,
        outbound_network_policy: Optional[CustomTaskOutboundNetworkPolicy] = None,
    ):
        """Create a custom task version without files from previous versions.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task
        base_environment_id: str
            the id of the base environment to use with the custom task version
        maximum_memory: Optional[int]
            A number in bytes about how much memory custom tasks' inference containers can run with.
        is_major_update: bool
            If the current version is 2.3, `True` would set the new version at `3.0`.
            `False` would set the new version at `2.4`.
            Defaults to `True`.
        folder_path: Optional[str]
            The path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path.
        required_metadata_values: Optional[List[RequiredMetadataValue]]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.
        outbound_network_policy: Optional[CustomTaskOutboundNetworkPolicy]
            You must enable custom task network access permissions to pass any value other than `None`!
            Specifies if you custom task version is able to make network calls. `None` will set the value
            to DataRobot's default.

        Returns
        -------
        CustomTaskVersion
            created custom task version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return cls._create(
            "post",
            custom_task_id,
            is_major_update,
            base_environment_id,
            folder_path,
            maximum_memory=maximum_memory,
            required_metadata_values=required_metadata_values,
            outbound_network_policy=outbound_network_policy,
        )

    @classmethod
    def create_from_previous(
        cls,
        custom_task_id: str,
        base_environment_id: str,
        maximum_memory: Optional[int] = None,
        is_major_update: bool = True,
        folder_path: Optional[str] = None,
        files_to_delete: Optional[list[str]] = None,
        required_metadata_values: Optional[List[RequiredMetadataValue]] = None,
        outbound_network_policy: Optional[CustomTaskOutboundNetworkPolicy] = None,
    ):
        """Create a custom task version containing files from a previous version.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task
        base_environment_id: str
            the id of the base environment to use with the custom task version
        maximum_memory: Optional[int]
            A number in bytes about how much memory custom tasks' inference containers can run with.
        is_major_update: bool
            If the current version is 2.3, `True` would set the new version at `3.0`.
            `False` would set the new version at `2.4`.
            Defaults to `True`.
        folder_path: Optional[str]
            The path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path.
        files_to_delete: Optional[List[str]]
            the list of a file items ids to be deleted
            Example: ["5ea95f7a4024030aba48e4f9", "5ea6b5da402403181895cc51"]
        required_metadata_values: Optional[List[RequiredMetadataValue]]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.
        outbound_network_policy: Optional[CustomTaskOutboundNetworkPolicy]
            You must enable custom task network access permissions to pass any value other than `None`!
            Specifies if you custom task version is able to make network calls. `None` will get the value
            from the previous version if you have the proper permissions or use DataRobot's default.

        Returns
        -------
        CustomTaskVersion
            created custom task version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return cls._create(
            "patch",
            custom_task_id,
            is_major_update,
            base_environment_id,
            folder_path,
            maximum_memory,
            files_to_delete=files_to_delete,
            required_metadata_values=required_metadata_values,
            outbound_network_policy=outbound_network_policy,
        )

    @classmethod
    def _create(  # pylint: disable=missing-function-docstring
        cls,
        method,
        custom_task_id,
        is_major_update,
        base_environment_id,
        folder_path=None,
        maximum_memory=None,
        files_to_delete=None,
        required_metadata_values=None,
        outbound_network_policy=None,
    ):
        url = cls._all_versions_path(custom_task_id)

        upload_data = [
            ("isMajorUpdate", str(is_major_update)),
            ("baseEnvironmentId", base_environment_id),
        ]
        if files_to_delete:
            upload_data += [("filesToDelete", file_id) for file_id in files_to_delete]

        if required_metadata_values is not None:
            upload_data.append(
                (
                    "requiredMetadataValues",
                    json.dumps(
                        [
                            {camelize(k): v for k, v in val.to_dict().items()}
                            for val in required_metadata_values
                        ]
                    ),
                )
            )

        if maximum_memory:
            upload_data.append(("maximumMemory", str(maximum_memory)))

        if outbound_network_policy is not None:
            upload_data.append(("outboundNetworkPolicy", outbound_network_policy.name))

        cls._verify_folder_path(folder_path)

        with contextlib.ExitStack() as stack:
            if folder_path:
                for dir_name, _, file_names in os.walk(folder_path):
                    for file_name in file_names:
                        file_path = os.path.join(dir_name, file_name)
                        file = stack.enter_context(open(file_path, "rb"))

                        upload_data.append(("file", (os.path.basename(file_path), file)))
                        upload_data.append(("filePath", os.path.relpath(file_path, folder_path)))

            encoder = MultipartEncoder(fields=upload_data)
            headers = {"Content-Type": encoder.content_type}
            response = cls._client.request(method, url, data=encoder, headers=headers)
        return cls.from_server_data(response.json())

    @staticmethod
    def _verify_folder_path(folder_path):
        if folder_path and not os.path.exists(folder_path):
            raise ValueError(f"The folder: {folder_path} does not exist.")

    @classmethod
    def list(cls, custom_task_id):
        """List custom task versions.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task

        Returns
        -------
        List[CustomTaskVersion]
            a list of custom task versions

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._all_versions_path(custom_task_id)
        data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_task_id, custom_task_version_id):
        """Get custom task version by id.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task
        custom_task_version_id: str
            the id of the custom task version to retrieve

        Returns
        -------
        CustomTaskVersion
            retrieved custom task version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._single_version_path(custom_task_id, custom_task_version_id)
        return cls.from_location(path)

    def download(self, file_path):
        """Download custom task version.

        .. versionadded:: v2.26

        Parameters
        ----------
        file_path: str
            path to create a file with custom task version content

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """

        response = self._client.get(
            self._single_version_path(self.custom_task_id, self.id) + "download/"
        )
        with open(file_path, "wb") as f:
            f.write(response.content)

    def update(self, description=None, required_metadata_values=None):
        """Update custom task version properties.

        .. versionadded:: v2.26

        Parameters
        ----------
        description: str
            new custom task version description
        required_metadata_values: List[RequiredMetadataValue]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {}
        if description:
            payload.update({"description": description})

        if required_metadata_values is not None:
            payload.update(
                {"requiredMetadataValues": [val.to_dict() for val in required_metadata_values]}
            )

        url = self._path.format(self.custom_task_id)
        path = f"{url}{self.id}/"

        response = self._client.patch(path, data=payload)

        data = response.json()
        new_version = CustomTaskVersion.from_server_data(data)
        self._update_values(new_version)

    def refresh(self):
        """Update custom task version with the latest data from server.

        .. versionadded:: v2.26

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.custom_task_id, self.id)
        self._update_values(new_object)

    def start_dependency_build(self):
        """Start the dependency build for a custom task version and return build status.
        .. versionadded:: v2.27

        Returns
        -------
        CustomTaskVersionDependencyBuild
            DTO of custom task version dependency build.
        """
        url = self._dependency_build_path.format(self.custom_task_id, self.id)
        response = self._client.post(url)
        server_data = response.json()
        return CustomTaskVersionDependencyBuild.from_server_data(server_data)

    def start_dependency_build_and_wait(self, max_wait):
        """Start the dependency build for a custom task version and wait while pulling status.
        .. versionadded:: v2.27

        Parameters
        ----------
        max_wait: int
            max time to wait for a build completion

        Returns
        -------
        CustomTaskVersionDependencyBuild
            DTO of custom task version dependency build.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        datarobot.errors.AsyncTimeoutError
            Raised if the dependency build is not finished after max_wait.
        """
        custom_task_id = self.custom_task_id
        custom_task_version_id = self.id

        def build_complete(response):
            server_data = response.json()
            if server_data["buildStatus"] in ["success", "failed"]:
                return CustomTaskVersionDependencyBuild.from_server_data(server_data)
            return None

        url = self._dependency_build_path.format(custom_task_id, custom_task_version_id)
        self._client.post(url)
        return wait_for_custom_resolution(self._client, url, build_complete, max_wait)

    def cancel_dependency_build(self):
        """Cancel custom task version dependency build that is in progress.
        .. versionadded:: v2.27

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        url = self._dependency_build_path.format(self.custom_task_id, self.id)
        self._client.delete(url)

    def get_dependency_build(self):
        """Retrieve information about a custom task version's dependency build.
        .. versionadded:: v2.27

        Returns
        -------
        CustomTaskVersionDependencyBuild
            DTO of custom task version dependency build.
        """

        url = self._dependency_build_path.format(self.custom_task_id, self.id)
        response = self._client.get(url)
        server_data = response.json()
        return CustomTaskVersionDependencyBuild.from_server_data(server_data)

    def download_dependency_build_log(self, file_directory="."):
        """Get log of a custom task version dependency build.
        .. versionadded:: v2.27

        Parameters
        ----------
        file_directory: str (optional, default is ".")
            Directory path where downloaded file is to save.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        url = self._dependency_build_log_path.format(self.custom_task_id, self.id)
        response = self._client.get(url)
        content_disposition = response.headers["Content-Disposition"]
        fiie_name = content_disposition[len("attachment; filename=") :]
        with open(os.path.join(file_directory, fiie_name), "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                f.write(chunk)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.models.execution_environment_version import ExecutionEnvironmentVersion
from datarobot.models.sharing import SharingAccess
from datarobot.utils.pagination import unpaginate


class RequiredMetadataKey(APIObject):
    """Definition of a metadata key that custom models using this environment must define

    .. versionadded:: v2.25

    Attributes
    ----------
    field_name: str
        The required field key. This value will be added as an environment
        variable when running custom models.
    display_name: str
        A human readable name for the required field.
    """

    _converter = t.Dict({t.Key("field_name"): String(), t.Key("display_name"): String()})

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return "{}(field_name={!r}, display_name={!r})".format(
            self.__class__.__name__,
            self.field_name,
            self.display_name,
        )

    def _set_values(self, field_name, display_name):
        self.field_name = field_name
        self.display_name = display_name

    def to_dict(self):
        return self._converter.check(
            {"field_name": self.field_name, "display_name": self.display_name}
        )


class ExecutionEnvironment(APIObject):
    """An execution environment entity.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        The ID of the execution environment.
    name: str
        The name of the execution environment.
    description: Optional[str]
        The description of the execution environment.
    programming_language: Optional[str]
        The programming language of the execution environment.
        Can be "python", "r", "java" or "other".
    is_public: Optional[bool]
        Public accessibility of environment, visible only for admin user.
    created_at: Optional[str]
        ISO-8601 formatted timestamp of when the execution environment version was created.
    latest_version: ExecutionEnvironmentVersion, optional
        The latest version of the execution environment.
    latest_successful_version: ExecutionEnvironmentVersion, optional
        The latest version of the execution environment, which contains a successfully built image.
    required_metadata_keys: Optional[List[RequiredMetadataKey]]
        The definition of metadata keys that custom models using this environment must define.
    use_cases: Optional[List[str]]
        A list of use-cases this environment may be used for.
    """

    _path = "executionEnvironments/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(max_length=255),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("programming_language", optional=True): String(),
            t.Key("is_public", optional=True): t.Bool(),
            t.Key("created", optional=True) >> "created_at": String(),
            t.Key("latest_version", optional=True): ExecutionEnvironmentVersion.schema,
            t.Key("latest_successful_version", optional=True): ExecutionEnvironmentVersion.schema,
            t.Key("required_metadata_keys", optional=True, default=list): t.List(
                RequiredMetadataKey.schema
            ),
            t.Key("use_cases", optional=True, default=list): t.List(String()),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id,
        name,
        description=None,
        programming_language=None,
        is_public=None,
        created_at=None,
        latest_version=None,
        latest_successful_version=None,
        required_metadata_keys=None,
        use_cases=None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.programming_language = programming_language
        self.is_public = is_public
        self.created_at = created_at
        self.latest_version = None
        self.latest_successful_version = None
        self.required_metadata_keys = [
            RequiredMetadataKey.from_data(key) for key in required_metadata_keys
        ]
        self.use_cases = use_cases

        if latest_version is not None:
            latest_version.pop("image_id", None)  # "image_id" is being removed in RAPTOR-2460
            self.latest_version = ExecutionEnvironmentVersion(**latest_version)

        if latest_successful_version is not None:
            latest_successful_version.pop("image_id", None)
            self.latest_successful_version = ExecutionEnvironmentVersion(
                **latest_successful_version
            )

    @classmethod
    def create(
        cls,
        name,
        description=None,
        programming_language=None,
        required_metadata_keys=None,
        is_public=None,
        use_cases=None,
    ):
        """Create an execution environment.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: str
            execution environment name
        description: Optional[str]
            execution environment description
        programming_language: Optional[str]
            programming language of the environment to be created.
            Can be "python", "r", "java" or "other". Default value - "other"
        required_metadata_keys: List[RequiredMetadataKey]
            Definition of a metadata keys that custom models using this environment must define
        is_public: bool, optional
            public accessibility of environment
        use_cases: List[str], optional
             List of use-cases this environment may be used for

        Returns
        -------
        ExecutionEnvironment
            created execution environment

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        required_metadata_keys = required_metadata_keys or []
        payload = {
            "name": name,
            "description": description,
            "programming_language": programming_language,
            "is_public": is_public,
            "use_cases": use_cases,
        }
        if required_metadata_keys:
            payload["required_metadata_keys"] = [key.to_dict() for key in required_metadata_keys]
        response = cls._client.post(cls._path, data=payload)
        environment_id = response.json()["id"]
        return cls.get(environment_id)

    @classmethod
    def list(
        cls,
        search_for=None,
        is_own: Optional[bool] = None,
        use_cases: Optional[str] = None,
        offset: Optional[int] = 0,
        limit: Optional[int] = 0,
    ):
        """List execution environments available to the user.

        .. versionadded:: v2.21

        Parameters
        ----------
        search_for: Optional[str]
            the string for filtering execution environment - only execution
            environments that contain the string in name or description will
            be returned.
        is_own: bool, optional
            Only return execution environments that were created by the current user.
        use_cases: str, optional
            Only return execution environments that contain the specified use case
        offset: Optional[int]
            The starting offset of the results. The default is 0.
        limit: Optional[int]
            The maximum number of objects to return. The default is 0 to maintain previous behavior.
            The default on the server is 20, with a maximum of 100.

        Returns
        -------
        List[ExecutionEnvironment]
            a list of execution environments.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        param: Dict[str, Any] = {"search_for": search_for}
        if is_own is not None:
            param["is_own"] = is_own
        if use_cases is not None:
            param["use_cases"] = use_cases

        if limit == 0:
            data = unpaginate(cls._path, param, cls._client)
        else:
            param["limit"] = limit
            param["offset"] = offset
            response = cls._client.get(cls._path, params=param).json()
            data = response["data"]

        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, execution_environment_id):
        """Get execution environment by its ID.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            ID of the execution environment to retrieve

        Returns
        -------
        ExecutionEnvironment
            retrieved execution environment

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = f"{cls._path}{execution_environment_id}/"
        return cls.from_location(path)

    def delete(self):
        """Delete execution environment.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = f"{self._path}{self.id}/"
        self._client.delete(url)

    def update(
        self,
        name=None,
        description=None,
        required_metadata_keys=None,
        is_public=None,
        use_cases=None,
    ):
        """Update execution environment properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        name: Optional[str]
            new execution environment name
        description: Optional[str]
            new execution environment description
        required_metadata_keys: List[RequiredMetadataKey]
            Definition of a metadata keys that custom models using this environment must define
        is_public: bool, optional
            public accessibility of environment
        use_cases: List[str], optional
             List of use-cases this environment may be used for

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {"name": name, "description": description}
        if required_metadata_keys is not None:
            payload["required_metadata_keys"] = [key.to_dict() for key in required_metadata_keys]
        if is_public is not None:
            payload["is_public"] = is_public
        if use_cases is not None:
            payload["use_cases"] = use_cases

        url = f"{self._path}{self.id}/"
        response = self._client.patch(url, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def refresh(self):
        """Update execution environment with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = f"{self._path}{self.id}/"
        response = self._client.get(url)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def get_access_list(self) -> List[SharingAccess]:
        """Retrieve access control settings of this environment.

        .. versionadded:: v2.36

        Returns
        -------
        list of :class:`SharingAccess <datarobot.SharingAccess>`
        """
        url = f"{self._path}{self.id}/accessControl/"
        return [
            SharingAccess.from_server_data(datum) for datum in unpaginate(url, {}, self._client)
        ]

    def share(self, access_list: List[SharingAccess]) -> None:
        """Update the access control settings of this execution environment.

        .. versionadded:: v2.36

        Parameters
        ----------
        access_list : list of :class:`SharingAccess <datarobot.ExecutionEnvironment>`
            A list of SharingAccess to update.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status

        Examples
        --------
        Transfer access to the execution environment from old_user@datarobot.com to new_user@datarobot.com

        .. code-block:: python

            import datarobot as dr

            new_access = dr.SharingAccess(new_user@datarobot.com,
                                          dr.enums.SHARING_ROLE.OWNER, can_share=True)
            access_list = [dr.SharingAccess(old_user@datarobot.com, None), new_access]

            dr.ExecutionEnvironment.get('environment-id').share(access_list)
        """
        payload = {
            "data": [access.collect_payload() for access in access_list],
        }
        nullable_query_params = {"role"}
        self._client.patch(
            f"{self._path}{self.id}/accessControl/",
            data=payload,
            keep_attrs=nullable_query_params,
        )



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional

import trafaret as t

from datarobot.enums import ModerationGuardModelTargetType
from datarobot.models.api_object import APIObject


class GuardModelInfo(APIObject):
    """
    Model information for moderation templates and configurations.
    Omitted optional values are stored and presented as:
    * []  (for class names)
    * None  (all others)
    """

    _converter = t.Dict(
        {
            t.Key("input_column_name"): t.String(),
            t.Key("output_column_name"): t.String(),
            t.Key("target_type"): t.Enum(*[e.value for e in ModerationGuardModelTargetType]),
            t.Key("replacement_text_column_name", optional=True): t.String(allow_blank=True),
            t.Key("class_names", optional=True): t.List(t.String()),
            t.Key("model_id", optional=True): t.String(allow_blank=True),
            t.Key("model_name", optional=True): t.String(allow_blank=True),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self: GuardModelInfo, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(target_type={!r})".format(
            self.__class__.__name__,
            self.target_type,
        )

    def _set_values(
        self: GuardModelInfo,
        input_column_name: str,
        output_column_name: str,
        target_type: ModerationGuardModelTargetType,
        replacement_text_column_name: Optional[str] = None,
        class_names: Optional[list[str]] = None,
        model_id: Optional[str] = None,
        model_name: Optional[str] = None,
    ) -> None:
        """
        Initialize a model information object based on values from init or server data.

        Parameters
        ----------
        input_column_name : str
            The name of the input column.
        output_column_name : str
            The name of the output column.
        target_type : ModerationGuardModelTargetType
            The target type of the model.
        replacement_text_column_name : Optional[str]
            The name of the replacement text column.
        class_names : list[str], optional
            The class names of the model.
        model_id : Optional[str]
            The ID of the model.
        model_name : Optional[str]
            The name of the model.
        """
        self.input_column_name = input_column_name
        self.output_column_name = output_column_name
        self.target_type = ModerationGuardModelTargetType(target_type)
        self.replacement_text_column_name = replacement_text_column_name
        self.class_names = class_names or []
        self.model_id = model_id
        self.model_name = model_name

    def to_dict(self: GuardModelInfo) -> Dict[str, str | List[str] | Optional[str]]:
        """
        Convert the model information object to a dictionary.
        """
        response = {
            "inputColumnName": self.input_column_name,
            "outputColumnName": self.output_column_name,
            "targetType": self.target_type.value,
            "classNames": self.class_names or [],
            "modelId": self.model_id,
            "modelName": self.model_name,
            "replacementTextColumnName": self.replacement_text_column_name,
        }
        return response

    @classmethod
    def ensure_object(cls, maybe_dict: Dict[str, str] | GuardModelInfo) -> GuardModelInfo:
        """intervention may arrive as an object, or as a dict. Return an object."""
        if isinstance(maybe_dict, GuardModelInfo):
            return maybe_dict
        elif isinstance(maybe_dict, dict):
            return GuardModelInfo(**maybe_dict)
        else:
            raise ValueError("Expected GuardModelInfo object or dictionary")



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Callable, Dict, NamedTuple, Optional, Set, Union
import warnings


class AppPlatformError(Exception):
    """
    Raised by :meth:`Client.request()` for requests that:
      - Return a non-200 HTTP response, or
      - Connection refused/timeout or
      - Response timeout or
      - Malformed request
      - Have a malformed/missing header in the response.
    """

    def __init__(
        self,
        exc_message: str,
        status_code: int,
        error_code: Optional[Union[str, int]] = None,
        json: Optional[Dict[Any, Any]] = None,
    ) -> None:
        super().__init__(exc_message)
        self.status_code = status_code
        self.error_code = error_code
        self.json = json or {}


class ServerError(AppPlatformError):
    """
    For 500-level responses from the server
    """


class ClientError(AppPlatformError):
    """
    For 400-level responses from the server
    has json parameter for additional information to be stored about error
    if need be
    """


class InputNotUnderstoodError(Exception):
    """
    Raised if a method is called in a way that cannot be understood
    """


class InvalidUsageError(Exception):
    """Raised when methods are called with invalid or incompatible arguments"""


class AllRetriesFailedError(Exception):
    """Raised when the retry manager does not successfully make a request"""


class InvalidModelCategoryError(Exception):
    """
    Raised when method specific for model category was called from wrong model
    """


class AsyncTimeoutError(Exception):
    """
    Raised when an asynchronous operation did not successfully get resolved
    within a specified time limit
    """


class AsyncFailureError(Exception):
    """
    Raised when querying an asynchronous status resulted in an exceptional
    status code (not 200 and not 303)
    """


class ProjectAsyncFailureError(AsyncFailureError):
    """
    When an AsyncFailureError occurs during project creation or finalizing the project
    settings for modeling. This exception will have the attributes ``status_code``
    indicating the unexpected status code from the server, and ``async_location`` indicating
    which asynchronous status object was being polled when the failure happened.
    """

    def __init__(self, exc_message: str, status_code: int, async_location: str) -> None:
        super().__init__(exc_message)
        self.status_code = status_code
        self.async_location = async_location


class AsyncProcessUnsuccessfulError(Exception):
    """
    Raised when querying an asynchronous status showed that async process
    was not successful
    """


class AsyncModelCreationError(Exception):
    """
    Raised when querying an asynchronous status showed that model creation
    was not successful
    """


class AsyncPredictionsGenerationError(Exception):
    """
    Raised when querying an asynchronous status showed that predictions
    generation was not successful
    """


class PendingJobFinished(Exception):
    """
    Raised when the server responds with a 303 for the pending creation of a
    resource.
    """


class JobNotFinished(Exception):
    """
    Raised when execution was trying to get a finished resource from a pending
    job, but the job is not finished
    """


class DuplicateFeaturesError(Exception):
    """
    Raised when trying to create featurelist with duplicating features
    """


class TrainingDataAssignmentError(Exception):
    """
    Raised when the training data assignment for a custom model version fails
    """

    def __init__(
        self, custom_model_id: str, custom_model_version_id: str, error_message: str
    ) -> None:
        self.custom_model_id = custom_model_id
        self.custom_model_version_id = custom_model_version_id
        self.error_message = error_message
        self.message = (
            f"Training data assignment failed for: "
            f"model ID: {custom_model_id}; "
            f"version ID: {custom_model_version_id}; "
            f"Error message: {error_message}"
        )
        super().__init__(self.message)


class DataRobotDeprecationWarning(DeprecationWarning):
    """
    Raised when using deprecated functions or using functions in a deprecated way

    See Also
    --------
    PlatformDeprecationWarning
    """


class IllegalFileName(Exception):
    """
    Raised when trying to use a filename we can't handle.
    """


class JobAlreadyRequested(ClientError):
    """
    Raised when the requested model has already been requested.
    """


class ContentRetrievalTerminatedError(Exception):
    """
    Raised when due to content retrieval error process of data retrieval was terminated.
    """


class UpdateAttributesError(AttributeError):
    def __init__(self, class_name: str, invalid_key: str, message: str = ""):
        self.class_name: str = class_name
        self.invalid_key: str = invalid_key
        self.message: str = message
        super().__init__(self.message)


class InvalidRatingTableWarning(Warning):
    """
    Raised when using interacting with rating tables that failed validation
    """


class PartitioningMethodWarning(Warning):
    """
    Raised when interacting with project methods related to partition classes, i.e.
    `Project.set_partitioning_method()` or `Project.set_datetime_partitioning()`.
    """


class NonPersistableProjectOptionWarning(Warning):
    """
    Raised when setting project options via `Project.set_options` if any of the options
    passed are not supported for POST requests to `/api/v2/project/{project_id}/options/`.
    All options that fall under this category can be found here:
    :meth:`datarobot.enums.NonPersistableProjectOptions`.
    """

    def __init__(self, options: Set[str]) -> None:
        self.message = f"""Project option(s) {options} will not be stored on the backend.
        The value set via this method will only be associated with a project instance for
        the duration of a client session. If you quit your session and reopen a new one before
        running autopilot, the value(s) of {options} will be lost.
        """
        super().__init__(self.message)


class OverwritingProjectOptionWarning(Warning):
    """
    Raised when setting project options via `Project.set_options` if any of the options
    passed have already been set to a value in `Project.advanced_options`, or if
    a different value is already stored in the endpoint `/api/v2/project/{project_id}/options/`.
    Precedence is given to the new value you passed in.
    """

    def __init__(self, options: Dict[str, NamedTuple]) -> None:
        details: Callable[[str], str] = (
            lambda option: f"{option} had a value of {options[option].old_value} "  # type: ignore[attr-defined]
            f"in the backend, possibly from previous calls to `Project.set_options`. {option} will be "
            f"overwritten to {options[option].new_value}. \n"
        )

        self.message = f"Project option(s) {list(options.keys())} are already configured and will be overwritten."
        for option in options:
            self.message += details(option)
        super().__init__(self.message)


class NoRedundancyImpactAvailable(Warning):
    """Raised when retrieving old feature impact data

    Redundancy detection was added in v2.13 of the API, and some projects, e.g. multiclass projects
    do not support redundancy detection. This warning is raised to make
    clear that redundancy detection is unavailable.
    """


class ParentModelInsightFallbackWarning(Warning):
    """
    Raised when insights are unavailable for a model and
    insight retrieval falls back to retrieving insights
    for a model's parent model
    """


class ProjectHasNoRecommendedModelWarning(Warning):
    """
    Raised when a project has no recommended model.
    """


class PlatformDeprecationWarning(Warning):
    """
    Raised when `Deprecation` header is returned in the API, for example a project may be
    deprecated as part of the 2022 Python 3 platform migration.

    See Also
    --------
    DataRobotDeprecationWarning
    """


class MultipleUseCasesNotAllowed(UserWarning):
    """
    Raised when a method decorated with add_to_use_case(allow_multiple=True) calls a method
    decorated with add_to_use_case(allow_multiple=False) with multiple UseCases passed
    """

    message = "Entity can't be added to multiple UseCases"


warnings.filterwarnings("default", category=DataRobotDeprecationWarning)
warnings.filterwarnings("always", category=PlatformDeprecationWarning)
warnings.filterwarnings("always", category=InvalidRatingTableWarning)
warnings.filterwarnings("always", category=ParentModelInsightFallbackWarning)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import TypedDict

from datarobot.models.api_object import APIObject
from datarobot.models.genai.chat_prompt import (
    Citation,
    citation_trafaret,
    confidence_scores_trafaret,
    ConfidenceScores,
    feedback_metadata_trafaret,
    FeedbackMetadata,
    FeedbackMetadataDict,
    result_metadata_trafaret,
    ResultMetadata,
)
from datarobot.models.genai.comparison_chat import ComparisonChat
from datarobot.models.genai.llm_blueprint import LLMBlueprint
from datarobot.utils import rawdict
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


def _get_genai_entity_id(entity: Union[ComparisonChat, ComparisonPrompt, LLMBlueprint, str]) -> str:
    """
    Get the entity ID from the entity parameter.

    Parameters
    ----------
    entity : ApiObject or str
        May be entity ID or the entity.

    Returns
    -------
    id : str
        The entity ID.
    """
    if isinstance(entity, str):
        return entity

    return entity.id


comparison_prompt_result_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("llm_blueprint_id"): t.String,
        t.Key("result_metadata", optional=True): t.Or(result_metadata_trafaret, t.Null),
        t.Key("result_text", optional=True): t.Or(t.String(allow_blank=True), t.Null),
        t.Key("confidence_scores", optional=True): t.Or(confidence_scores_trafaret, t.Null),
        t.Key("citations"): t.List(citation_trafaret),
        t.Key("execution_status"): t.String,
        t.Key("chat_context_id", optional=True): t.Or(t.String, t.Null),
        t.Key("comparison_prompt_result_ids_included_in_history", optional=True): t.Or(
            t.List(t.String), t.Null
        ),
    }
).ignore_extra("*")


comparison_prompt_trafaret = t.Dict(
    {
        t.Key("id"): t.String,
        t.Key("text"): t.String(allow_blank=True),
        t.Key("results"): t.List(comparison_prompt_result_trafaret),
        t.Key("creation_date"): t.String,
        t.Key("creation_user_id"): t.String,
        t.Key("comparison_chat_id", optional=True): t.Or(t.String, t.Null),
    }
).ignore_extra("*")


feedback_result_trafaret = t.Dict(
    {
        t.Key("comparison_prompt_result_id"): t.String,
        t.Key("feedback_metadata"): feedback_metadata_trafaret,
    }
).ignore_extra("*")


class FeedbackResultDict(TypedDict):
    comparison_prompt_result_id: str
    feedback_metadata: FeedbackMetadataDict


class FeedbackResult(APIObject):
    """
    Feedback associated with a comparison prompt result.

    Attributes
    ----------
    comparison_prompt_result_id : str
        The ID of the comparison prompt result associated with the feedback.
    feedback_metadata : FeedbackMetadata
        The metadata for the feedback.
    """

    _converter = feedback_result_trafaret

    def __init__(self, comparison_prompt_result_id: str, feedback_metadata: FeedbackMetadata):
        self.comparison_prompt_result_id = comparison_prompt_result_id
        self.feedback_metadata = feedback_metadata

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(comparison_prompt_result_id={self.comparison_prompt_result_id},"
            f"feedback_metadata={self.feedback_metadata})"
        )

    def to_dict(self) -> FeedbackResultDict:
        return {
            "comparison_prompt_result_id": self.comparison_prompt_result_id,
            "feedback_metadata": self.feedback_metadata.to_dict(),
        }


class ComparisonPromptResult(APIObject):
    """
    Metadata for a DataRobot GenAI comparison prompt result.

    Attributes
    ----------
    id: str
        The ID of the comparison prompt result.
    llm_blueprint_id : str
        The ID of the LLM blueprint associated with the chat prompt.
    result_metadata : ResultMetadata or None
        Metadata for the result of the chat prompt submission.
    result_text: str or None
        The result text from the chat prompt submission.
    confidence_scores: ConfidenceScores or None
        The confidence scores if there is a vector database associated with the chat prompt.
    citations: list[Citation]
        List of citations from text retrieved from the vector database, if any.
    execution_status: str
        The execution status of the chat prompt.
    chat_context_id: Optional[str], optional
        The ID of the chat context for this comparison prompt result.
    comparison_prompt_result_ids_included_in_history: Optional[List[str]], optional
        The IDs of the comparison prompt results included in the chat history for this
        comparison prompt result.
    """

    _converter = comparison_prompt_result_trafaret

    def __init__(
        self,
        id: str,
        llm_blueprint_id: str,
        citations: List[Dict[str, Any]],
        execution_status: str,
        result_metadata: Optional[Dict[str, Any]] = None,
        result_text: Optional[str] = None,
        confidence_scores: Optional[Dict[str, float]] = None,
        chat_context_id: Optional[str] = None,
        comparison_prompt_result_ids_included_in_history: Optional[List[str]] = None,
    ):
        self.id = id
        self.llm_blueprint_id = llm_blueprint_id
        self.citations = [Citation.from_server_data(citation) for citation in citations]
        self.execution_status = execution_status
        self.result_metadata = (
            ResultMetadata.from_server_data(result_metadata) if result_metadata else None
        )
        self.result_text = result_text
        self.confidence_scores = (
            ConfidenceScores.from_server_data(confidence_scores) if confidence_scores else None
        )
        self.chat_context_id = chat_context_id
        self.comparison_prompt_result_ids_included_in_history = (
            comparison_prompt_result_ids_included_in_history
        )
        self.citations = [Citation.from_server_data(citation) for citation in citations]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(llm_blueprint_id={self.llm_blueprint_id}, "
            f"execution_status={self.execution_status})"
        )


class ComparisonPrompt(APIObject):
    """
    Metadata for a DataRobot GenAI comparison prompt.

    Attributes
    ----------
    id : str
        Comparison prompt ID.
    text : str
        The prompt text.
    results : list[ComparisonPromptResult]
        The list of results for individual LLM blueprints that are part of the comparison prompt.
    creation_date : str
        The date when the playground was created.
    creation_user_id : str
        ID of the creating user.
    comparison_chat_id : str
        The ID of the comparison chat this comparison prompt is associated with.
    metadata_filter: Optional[Dict[str, Any] | None]
        The metadata filter to apply to the vector database.

        Supports:
        - None or empty dict (no filters): Considers all documents
        - Multiple field filters (implicit AND): {"a": 1, "b": "b"}
        - Comparison operators: {"field": {"$gt": 5}}
        - Logical operators: {"$and": [...], "$or": [...]}
        - Nested combinations of the above

        Comparison operators:
        - $eq: equal to (string, int, float, bool)
        - $ne: not equal to (string, int, float, bool)
        - $gt: greater than (int, float)
        - $gte: greater than or equal to (int, float)
        - $lt: less than (int, float)
        - $lte: less than or equal to (int, float)
        - $in: a value is in list (string, int, float, bool)
        - $nin: a value is not in list (string, int, float, bool)
        - $contains: a string contains a value (string)
        - $not_contains: a string does not contain a value (string)
    """

    _path = "api/v2/genai/comparisonPrompts"

    _converter = comparison_prompt_trafaret

    def __init__(
        self,
        id: str,
        text: str,
        results: List[Dict[str, Any]],
        creation_date: str,
        creation_user_id: str,
        comparison_chat_id: Optional[str] = None,
        metadata_filter: Optional[Dict[str, Any]] = None,
    ):
        self.id = id
        self.text = text
        self.results = [ComparisonPromptResult.from_server_data(result) for result in results]
        self.creation_date = creation_date
        self.creation_user_id = creation_user_id
        self.comparison_chat_id = comparison_chat_id
        self.results = [ComparisonPromptResult.from_server_data(result) for result in results]
        self.metadata_filter = metadata_filter

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, text={self.text[:1000]})"

    def update(
        self,
        additional_llm_blueprints: Optional[List[Union[LLMBlueprint, str]]] = None,
        wait_for_completion: bool = False,
        feedback_result: Optional[FeedbackResult] = None,
        **kwargs: Any,
    ) -> ComparisonPrompt:
        """
        Update the comparison prompt.

        Parameters
        ----------
        additional_llm_blueprints : list[LLMBlueprint or str]
            The additional LLM blueprints you want to submit the comparison prompt.

        Returns
        -------
        comparison_prompt : ComparisonPrompt
            The updated comparison prompt.
        """
        payload = {
            "additionalLLMBlueprintIds": (
                [_get_genai_entity_id(bp) for bp in additional_llm_blueprints]
                if additional_llm_blueprints
                else None
            ),
            "feedbackResult": feedback_result.to_dict() if feedback_result else None,
        }
        url = f"{self._client.domain}/{self._path}/{_get_genai_entity_id(self.id)}/"
        r_data = self._client.patch(url, data=payload)
        if wait_for_completion and additional_llm_blueprints:
            # If no additional_llm_blueprints then we get no location header
            location = wait_for_async_resolution(self._client, r_data.headers["Location"])
            return self.from_location(location)
        else:
            # Update route returns empty string so we need to GET here
            r_data = self._client.get(url)
            return self.from_server_data(r_data.json())

    @classmethod
    def create(
        cls,
        llm_blueprints: List[Union[LLMBlueprint, str]],
        text: str,
        comparison_chat: Optional[Union[ComparisonChat, str]] = None,
        wait_for_completion: bool = False,
        metadata_filter: Optional[Dict[str, Any]] = None,
    ) -> ComparisonPrompt:
        """
        Create a new ComparisonPrompt. This submits the prompt text to the LLM blueprints that
        are specified.

        Parameters
        ----------
        llm_blueprints : list[LLMBlueprint or str]
            The LLM blueprints associated with the created comparison prompt.
            Accepts LLM blueprints or IDs.
        text : str
            The prompt text.
        comparison_chat: Optional[ComparisonChat or str], optional
            The comparison chat to add the comparison prompt to. Accepts `ComparisonChat` or
            comparison chat ID.
        wait_for_completion : bool
            If set to True code will wait for the chat prompt job to complete before
            returning the result (up to 10 minutes, raising timeout error after that).
            Otherwise, you can check current status by using ChatPrompt.get with returned ID.
        metadata_filter: Optional[Dict[str, Any] | None]
            The metadata filter to apply to the vector database.

            Supports:
            - None or empty dict (no filters): Considers all documents
            - Multiple field filters (implicit AND): {"a": 1, "b": "b"}
            - Comparison operators: {"field": {"$gt": 5}}
            - Logical operators: {"$and": [...], "$or": [...]}
            - Nested combinations of the above

            Comparison operators:
            - $eq: equal to (string, int, float, bool)
            - $ne: not equal to (string, int, float, bool)
            - $gt: greater than (int, float)
            - $gte: greater than or equal to (int, float)
            - $lt: less than (int, float)
            - $lte: less than or equal to (int, float)
            - $in: a value is in list (string, int, float, bool)
            - $nin: a value is not in list (string, int, float, bool)
            - $contains: a string contains a value (string)
            - $not_contains: a string does not contain a value (string)

        Returns
        -------
        comparison_prompt : ComparisonPrompt
            The created comparison prompt.
        """
        payload = {
            "llm_blueprint_ids": [
                _get_genai_entity_id(llm_blueprint) for llm_blueprint in llm_blueprints
            ],
            "comparison_chat_id": (
                _get_genai_entity_id(comparison_chat) if comparison_chat else None
            ),
            "text": text,
            "metadata_filter": rawdict(metadata_filter) if metadata_filter else None,
        }

        url = f"{cls._client.domain}/{cls._path}/"
        r_data = cls._client.post(url, data=payload)
        if wait_for_completion:
            location = wait_for_async_resolution(cls._client, r_data.headers["Location"])
            return cls.from_location(location)
        return cls.from_server_data(r_data.json())

    @classmethod
    def get(cls, comparison_prompt: Union[ComparisonPrompt, str]) -> ComparisonPrompt:
        """
        Retrieve a single comparison prompt.

        Parameters
        ----------
        comparison_prompt : str
            The comparison prompt you want to retrieve. Accepts entity or ID.

        Returns
        -------
        comparison_prompt : ComparisonPrompt
            The requested comparison prompt.
        """
        url = f"{cls._client.domain}/{cls._path}/{_get_genai_entity_id(comparison_prompt)}/"
        r_data = cls._client.get(url)
        return cls.from_server_data(r_data.json())

    @classmethod
    def list(
        cls,
        llm_blueprints: Optional[List[Union[LLMBlueprint, str]]] = None,
        comparison_chat: Optional[Union[ComparisonChat, str]] = None,
    ) -> List[ComparisonPrompt]:
        """
        List all comparison prompts available to the user that include the specified LLM blueprints
        or from the specified comparison chat.

        Parameters
        ----------
        llm_blueprints : Optional[List[Union[LLMBlueprint, str]]], optional
            The returned comparison prompts are only those associated with the specified LLM
            blueprints. Accepts either `LLMBlueprint` or LLM blueprint ID.
        comparison_chat : Optional[Union[ComparisonChat, str]], optional
            The returned comparison prompts are only those associated with the specified comparison
            chat. Accepts either `ComparisonChat` or comparison chat ID.

        Returns
        -------
        comparison_prompts : list[ComparisonPrompt]
            A list of comparison prompts available to the user that use the specified LLM
            blueprints.
        """
        llm_blueprint_ids = (
            [_get_genai_entity_id(llm_blueprint) for llm_blueprint in llm_blueprints]
            if llm_blueprints
            else None
        )
        comparison_chat_id = _get_genai_entity_id(comparison_chat) if comparison_chat else None
        params = {
            "llm_blueprint_ids": llm_blueprint_ids,
            "comparison_chat_id": comparison_chat_id,
        }
        url = f"{cls._client.domain}/{cls._path}/"
        r_data = unpaginate(url, params, cls._client)
        return [cls.from_server_data(data) for data in r_data]

    def delete(self) -> None:
        """
        Delete the single comparison prompt.
        """
        url = f"{self._client.domain}/{self._path}/{_get_genai_entity_id(self.id)}/"
        self._client.delete(url)
